# Copyright Anne M. Archibald 2008
# Released under the scipy license
import numpy as np
cimport numpy as np
cimport libc.stdlib as stdlib
cimport cython

import kdtree

cdef extern from "stdlib.h":
    # NOTE that size_t might not be int
    void *alloca(int)

cdef np.float64_t infinity = np.inf

__all__ = ['cKDTree']

# priority queue
cdef union heapcontents:
    int intdata
    char* ptrdata

cdef struct heapitem:
    np.float64_t priority
    heapcontents contents

cdef struct heap:
    int n
    heapitem* heap
    int space

cdef inline heapcreate(heap* self,int initial_size):
    self.space = initial_size
    self.heap = <heapitem*>stdlib.malloc(sizeof(heapitem)*self.space)
    self.n=0

cdef inline heapdestroy(heap* self):
    stdlib.free(self.heap)

cdef inline heapresize(heap* self, int new_space):
    if new_space<self.n:
        raise ValueError("Heap containing %d items cannot be resized to %d" % (self.n, new_space))
    self.space = new_space
    self.heap = <heapitem*>stdlib.realloc(<void*>self.heap,new_space*sizeof(heapitem))

cdef inline heappush(heap* self, heapitem item):
    cdef int i
    cdef heapitem t

    self.n += 1
    if self.n>self.space:
        heapresize(self,2*self.space+1)

    i = self.n-1
    self.heap[i] = item
    while i>0 and self.heap[i].priority<self.heap[(i-1)//2].priority:
        t = self.heap[(i-1)//2]
        self.heap[(i-1)//2] = self.heap[i]
        self.heap[i] = t
        i = (i-1)//2

cdef heapitem heappeek(heap* self):
    return self.heap[0]

cdef heapremove(heap* self):
    cdef heapitem t
    cdef int i, j, k, l

    self.heap[0] = self.heap[self.n-1]
    self.n -= 1
    if self.n < self.space//4 and self.space>40: #FIXME: magic number
        heapresize(self,self.space//2+1)

    i=0
    j=1
    k=2
    while ((j<self.n and 
                self.heap[i].priority > self.heap[j].priority or
            k<self.n and 
                self.heap[i].priority > self.heap[k].priority)):
        if k<self.n and self.heap[j].priority>self.heap[k].priority:
            l = k
        else:
            l = j
        t = self.heap[l]
        self.heap[l] = self.heap[i]
        self.heap[i] = t
        i = l
        j = 2*i+1
        k = 2*i+2

cdef heapitem heappop(heap* self):
    cdef heapitem it
    it = heappeek(self)
    heapremove(self)
    return it





# utility functions
cdef inline np.float64_t dmax(np.float64_t x, np.float64_t y):
    if x>y:
        return x
    else:
        return y
cdef inline np.float64_t dabs(np.float64_t x):
    if x>0:
        return x
    else:
        return -x
cdef inline np.float64_t dmin(np.float64_t x, np.float64_t y):
    if x<y:
        return x
    else:
        return y
cdef inline np.float64_t _distance_p(np.float64_t*x,np.float64_t*y,np.float64_t p,int k,np.float64_t upperbound,
    np.float64_t*period):
    """Compute the distance between x and y

    Computes the Minkowski p-distance to the power p between two points.
    If the distance**p is larger than upperbound, then any number larger
    than upperbound may be returned (the calculation is truncated).
    
    Periodicity added by S. Skory.
    """
    cdef int i
    cdef np.float64_t r, m
    r = 0
    if p==infinity:
        for i in range(k):
            m = dmin(dabs(x[i] - y[i]), period[i] - dabs(x[i] - y[i]))
            r = dmax(r,m)
            if r>upperbound:
                return r
    elif p==1:
        for i in range(k):
            m = dmin(dabs(x[i] - y[i]), period[i] - dabs(x[i] - y[i]))
            r += m
            if r>upperbound:
                return r
    elif p==2:
        for i in range(k):
            m = dmin(dabs(x[i] - y[i]), period[i] - dabs(x[i] - y[i]))
            r += m*m
            if r>upperbound:
                return r
    else:
        for i in range(k):
            m = dmin(dabs(x[i] - y[i]), period[i] - dabs(x[i] - y[i]))
            r += m**p
            if r>upperbound:
                return r
    return r



# Tree structure
cdef struct innernode:
    int split_dim
    int n_points
    np.float64_t split
    np.float64_t* maxes
    np.float64_t* mins
    innernode* less
    innernode* greater
cdef struct leafnode:
    int split_dim
    int n_points
    int start_idx
    int end_idx
    np.float64_t* maxes
    np.float64_t* mins

# this is the standard trick for variable-size arrays:
# malloc sizeof(nodeinfo)+self.m*sizeof(np.float64_t) bytes.
cdef struct nodeinfo:
    innernode* node
    np.float64_t side_distances[0]

cdef class cKDTree:
    """kd-tree for quick nearest-neighbor lookup

    This class provides an index into a set of k-dimensional points
    which can be used to rapidly look up the nearest neighbors of any
    point. 

    The algorithm used is described in Maneewongvatana and Mount 1999. 
    The general idea is that the kd-tree is a binary trie, each of whose
    nodes represents an axis-aligned hyperrectangle. Each node specifies
    an axis and splits the set of points based on whether their coordinate
    along that axis is greater than or less than a particular value. 

    During construction, the axis and splitting point are chosen by the 
    "sliding midpoint" rule, which ensures that the cells do not all
    become long and thin. 

    The tree can be queried for the r closest neighbors of any given point 
    (optionally returning only those within some maximum distance of the 
    point). It can also be queried, with a substantial gain in efficiency, 
    for the r approximate closest neighbors.

    For large dimensions (20 is already large) do not expect this to run 
    significantly faster than brute force. High-dimensional nearest-neighbor
    queries are a substantial open problem in computer science.

    Parameters
    ----------
    data : array-like, shape (n,m)
        The n data points of dimension m to be indexed. This array is 
        not copied unless this is necessary to produce a contiguous 
        array of np.float64_ts, and so modifying this data will result in 
        bogus results.
    leafsize : positive integer
        The number of points at which the algorithm switches over to
        brute-force.

    """

    cdef innernode* tree 
    cdef readonly object data
    cdef np.float64_t* raw_data
    cdef readonly int n, m
    cdef readonly int leafsize
    cdef readonly object maxes
    cdef np.float64_t* raw_maxes
    cdef readonly object mins
    cdef np.float64_t* raw_mins
    cdef object indices
    cdef np.int64_t* raw_indices
    def __init__(cKDTree self, data, int leafsize=10):
        cdef np.ndarray[np.float64_t, ndim=2] inner_data
        cdef np.ndarray[np.float64_t, ndim=1] inner_maxes
        cdef np.ndarray[np.float64_t, ndim=1] inner_mins
        cdef np.ndarray[np.int64_t, ndim=1] inner_indices
        self.data = np.ascontiguousarray(data,dtype="float64")
        self.n, self.m = np.shape(self.data)
        self.leafsize = leafsize
        if self.leafsize<1:
            raise ValueError("leafsize must be at least 1")
        self.maxes = np.ascontiguousarray(np.amax(self.data,axis=0))
        self.mins = np.ascontiguousarray(np.amin(self.data,axis=0))
        self.indices = np.ascontiguousarray(np.arange(self.n,dtype=np.int64))

        inner_data = self.data
        self.raw_data = <np.float64_t*>inner_data.data
        inner_maxes = self.maxes
        self.raw_maxes = <np.float64_t*>inner_maxes.data
        inner_mins = self.mins
        self.raw_mins = <np.float64_t*>inner_mins.data
        inner_indices = self.indices
        self.raw_indices = <np.int64_t*>inner_indices.data

        self.tree = self.__build(0, self.n, self.raw_maxes, self.raw_mins)

    cdef innernode* __build(cKDTree self, int start_idx, int end_idx, np.float64_t* maxes, np.float64_t* mins):
        cdef leafnode* n
        cdef innernode* ni
        cdef int i, j, t, p, q, d
        cdef np.float64_t size, split, minval, maxval
        cdef np.float64_t*mids
        if end_idx-start_idx<=self.leafsize:
            n = <leafnode*>stdlib.malloc(sizeof(leafnode))
            # Skory
            n.maxes = <np.float64_t*>stdlib.malloc(sizeof(np.float64_t)*self.m)
            n.mins = <np.float64_t*>stdlib.malloc(sizeof(np.float64_t)*self.m)
            for i in range(self.m):
                n.maxes[i] = maxes[i]
                n.mins[i] = mins[i]
            n.split_dim = -1
            n.start_idx = start_idx
            n.end_idx = end_idx
            return <innernode*>n
        else:
            d = 0 
            size = 0
            for i in range(self.m):
                if maxes[i]-mins[i] > size:
                    d = i
                    size =  maxes[i]-mins[i]
            maxval = maxes[d]
            minval = mins[d]
            if maxval==minval:
                # all points are identical; warn user?
                n = <leafnode*>stdlib.malloc(sizeof(leafnode))
                n.split_dim = -1
                n.start_idx = start_idx
                n.end_idx = end_idx
                return <innernode*>n

            split = (maxval+minval)/2

            p = start_idx
            q = end_idx-1
            while p<=q:
                if self.raw_data[self.raw_indices[p]*self.m+d]<split:
                    p+=1
                elif self.raw_data[self.raw_indices[q]*self.m+d]>=split:
                    q-=1
                else:
                    t = self.raw_indices[p]
                    self.raw_indices[p] = self.raw_indices[q]
                    self.raw_indices[q] = t
                    p+=1
                    q-=1

            # slide midpoint if necessary
            if p==start_idx:
                # no points less than split
                j = start_idx
                split = self.raw_data[self.raw_indices[j]*self.m+d]
                for i in range(start_idx+1, end_idx):
                    if self.raw_data[self.raw_indices[i]*self.m+d]<split:
                        j = i
                        split = self.raw_data[self.raw_indices[j]*self.m+d]
                t = self.raw_indices[start_idx]
                self.raw_indices[start_idx] = self.raw_indices[j]
                self.raw_indices[j] = t
                p = start_idx+1
                q = start_idx
            elif p==end_idx:
                # no points greater than split
                j = end_idx-1
                split = self.raw_data[self.raw_indices[j]*self.m+d]
                for i in range(start_idx, end_idx-1):
                    if self.raw_data[self.raw_indices[i]*self.m+d]>split:
                        j = i
                        split = self.raw_data[self.raw_indices[j]*self.m+d]
                t = self.raw_indices[end_idx-1]
                self.raw_indices[end_idx-1] = self.raw_indices[j]
                self.raw_indices[j] = t
                p = end_idx-1
                q = end_idx-2

            # construct new node representation
            ni = <innernode*>stdlib.malloc(sizeof(innernode))

            mids = <np.float64_t*>stdlib.malloc(sizeof(np.float64_t)*self.m)
            for i in range(self.m):
                mids[i] = maxes[i]
            mids[d] = split
            ni.less = self.__build(start_idx,p,mids,mins)

            for i in range(self.m):
                mids[i] = mins[i]
            mids[d] = split
            ni.greater = self.__build(p,end_idx,maxes,mids)

            stdlib.free(mids)

            ni.split_dim = d
            ni.split = split
            # Skory
            ni.maxes = <np.float64_t*>stdlib.malloc(sizeof(np.float64_t)*self.m)
            ni.mins = <np.float64_t*>stdlib.malloc(sizeof(np.float64_t)*self.m)
            for i in range(self.m):
                ni.maxes[i] = maxes[i]
                ni.mins[i] = mins[i]

            return ni
                    
    cdef __free_tree(cKDTree self, innernode* node):
        if node.split_dim!=-1:
            self.__free_tree(node.less)
            self.__free_tree(node.greater)
        stdlib.free(node.maxes) # Skory
        stdlib.free(node.mins)
        stdlib.free(node)

    def __dealloc__(cKDTree self):
        if <int>(self.tree) == 0:
            # should happen only if __init__ was never called
            return
        self.__free_tree(self.tree)

    cdef void __query(cKDTree self, 
            np.float64_t*result_distances, 
            np.int64_t*result_indices, 
            np.float64_t*x, 
            int k, 
            np.float64_t eps, 
            np.float64_t p, 
            np.float64_t distance_upper_bound,
            np.float64_t*period):
        assert(p == 2)
        assert(eps == 0.0)
        assert(distance_upper_bound == infinity)
        cdef heap q
        cdef heap neighbors

        cdef int i, j, i2, j2
        cdef np.float64_t t, y
        cdef nodeinfo* inf
        cdef nodeinfo* inf2
        cdef np.float64_t d, di
        cdef np.float64_t m_left, m_right, m
        cdef np.float64_t epsfac
        cdef np.float64_t min_distance
        cdef np.float64_t far_min_distance
        cdef heapitem it, it2, neighbor
        cdef leafnode* node
        cdef innernode* inode
        cdef innernode* near
        cdef innernode* far
        cdef np.float64_t* side_distances

        # priority queue for chasing nodes
        # entries are:
        #  minimum distance between the cell and the target
        #  distances between the nearest side of the cell and the target
        #  the head node of the cell
        heapcreate(&q,12)

        # priority queue for the nearest neighbors
        # furthest known neighbor first
        # entries are (-distance**p, i)
        heapcreate(&neighbors,k)

        # set up first nodeinfo
        inf = <nodeinfo*>stdlib.malloc(sizeof(nodeinfo)+self.m*sizeof(np.float64_t)) 
        inf.node = self.tree
        for i in range(self.m):
            inf.side_distances[i] = 0
            t = x[i]-self.raw_maxes[i]
            if t>inf.side_distances[i]:
                inf.side_distances[i] = t
            else:
                t = self.raw_mins[i]-x[i]
                if t>inf.side_distances[i]:
                    inf.side_distances[i] = t
            inf.side_distances[i]=inf.side_distances[i]*inf.side_distances[i]

        # compute first distance
        min_distance = 0.
        for i in range(self.m):
            min_distance += inf.side_distances[i]

        # fiddle approximation factor
        epsfac=1

        while True:
            if inf.node.split_dim==-1:
                node = <leafnode*>inf.node

                # brute-force
                for i in range(node.start_idx,node.end_idx):
                    d = 0.0
                    for i2 in range(self.m):
                        y = self.raw_data[self.raw_indices[i]*self.m + i2]
                        di = dmin(dabs(x[i2] - y), period[i2] - dabs(x[i2] - y))
                        d += di*di
                    if d<distance_upper_bound:
                        # replace furthest neighbor
                        if neighbors.n==k:
                            heapremove(&neighbors)
                        neighbor.priority = -d
                        neighbor.contents.intdata = self.raw_indices[i]
                        heappush(&neighbors,neighbor)

                        # adjust upper bound for efficiency
                        if neighbors.n==k:
                            distance_upper_bound = -heappeek(&neighbors).priority
                # done with this node, get another
                stdlib.free(inf)
                if q.n==0:
                    # no more nodes to visit
                    break
                else:
                    it = heappop(&q)
                    inf = <nodeinfo*>it.contents.ptrdata
                    min_distance = it.priority
            else:
                inode = <innernode*>inf.node

                # we don't push cells that are too far onto the queue at all,
                # but since the distance_upper_bound decreases, we might get 
                # here even if the cell's too far
                if min_distance>distance_upper_bound*epsfac:
                    # since this is the nearest cell, we're done, bail out
                    stdlib.free(inf)
                    # free all the nodes still on the heap
                    for i in range(q.n):
                        stdlib.free(q.heap[i].contents.ptrdata)
                    break

                # set up children for searching
                if x[inode.split_dim]<inode.split:
                    near = inode.less
                    far = inode.greater
                else:
                    near = inode.greater
                    far = inode.less

                # near child is at the same distance as the current node
                # we're going here next, so no point pushing it on the queue
                # no need to recompute the distance or the side_distances
                inf.node = near

                # far child is further by an amount depending only
                # on the split value; compute its distance and side_distances
                # and push it on the queue if it's near enough
                inf2 = <nodeinfo*>stdlib.malloc(sizeof(nodeinfo)+self.m*sizeof(np.float64_t)) 
                it2.contents.ptrdata = <char*> inf2
                inf2.node = far

                # Periodicity added by S Skory
                m_left = dmin( dabs(far.mins[inode.split_dim] - x[inode.split_dim]), \
                    period[inode.split_dim] -  dabs(far.mins[inode.split_dim] - x[inode.split_dim]))
                m_right = dmin( dabs(far.maxes[inode.split_dim] - x[inode.split_dim]), \
                    period[inode.split_dim] -  dabs(far.maxes[inode.split_dim] - x[inode.split_dim]))
                m = dmin(m_left,m_right)

                # most side distances unchanged
                for i in range(self.m):
                    inf2.side_distances[i] = inf.side_distances[i]

                # one side distance changes
                # we can adjust the minimum distance without recomputing
                inf2.side_distances[inode.split_dim] = m*m
                #far_min_distance = min_distance - inf.side_distances[inode.split_dim] + inf2.side_distances[inode.split_dim]
                far_min_distance = m*m

                it2.priority = far_min_distance


                # far child might be too far, if so, don't bother pushing it
                if far_min_distance<=distance_upper_bound*epsfac:
                    heappush(&q,it2)
                else:
                    stdlib.free(inf2)
                    # just in case
                    it2.contents.ptrdata = <char*> 0

        # fill output arrays with sorted neighbors 
        for i in range(neighbors.n-1,-1,-1):
            neighbor = heappop(&neighbors) # FIXME: neighbors may be realloced
            result_indices[i] = neighbor.contents.intdata
            result_distances[i] = (-neighbor.priority) #**(1./p) S. Skory

        heapdestroy(&q)
        heapdestroy(&neighbors)

    def query(cKDTree self, object x, int k=1, np.float64_t eps=0, np.float64_t p=2, 
            np.float64_t distance_upper_bound=infinity, object period=None):
        """query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf,
           period=None)
        
        Query the kd-tree for nearest neighbors.

        Parameters
        ----------
        x : array_like, last dimension self.m
            An array of points to query.
        k : int
            The number of nearest neighbors to return.
        eps : non-negative float
            Return approximate nearest neighbors; the k-th returned value 
            is guaranteed to be no further than (1 + `eps`) times the 
            distance to the real k-th nearest neighbor.
        p : float, 1 <= p <= infinity
            Which Minkowski p-norm to use. 
            1 is the sum-of-absolute-values "Manhattan" distance.
            2 is the usual Euclidean distance.
            infinity is the maximum-coordinate-difference distance.
        distance_upper_bound : non-negative float
            Return only neighbors within this distance.  This is used to prune
            tree searches, so if you are doing a series of nearest-neighbor
            queries, it may help to supply the distance to the nearest neighbor
            of the most recent point.

        Returns
        -------
        d : ndarray of floats
            The distances to the nearest neighbors. 
            If `x` has shape tuple+(self.m,), then `d` has shape tuple+(k,).
            Missing neighbors are indicated with infinite distances.
        i : ndarray of ints
            The locations of the neighbors in self.data.
            If `x` has shape tuple+(self.m,), then `i` has shape tuple+(k,).
            Missing neighbors are indicated with self.n.

        """
        cdef np.ndarray[np.int64_t, ndim=2] ii
        cdef np.ndarray[np.float64_t, ndim=2] dd
        cdef np.ndarray[np.float64_t, ndim=2] xx
        cdef np.ndarray[np.float64_t, ndim=1] cperiod
        cdef int c
        x = np.asarray(x).astype("float64")
        if period is None:
            period = np.array([np.inf]*self.m)
        else:
            period = np.asarray(period).astype("float64")
        cperiod = np.ascontiguousarray(period)
        if np.shape(x)[-1] != self.m:
            raise ValueError("x must consist of vectors of length %d but has shape %s" % (self.m, np.shape(x)))
        if p<1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")
        if len(x.shape)==1:
            single = True
            x = x[np.newaxis,:]
        else:
            single = False
        retshape = np.shape(x)[:-1]
        n = np.prod(retshape)
        xx = np.reshape(x,(n,self.m))
        xx = np.ascontiguousarray(xx)
        dd = np.empty((n,k),dtype="float64")
        dd.fill(infinity)
        ii = np.empty((n,k),dtype="int64")
        ii.fill(self.n)
        for c in range(n):
            self.__query(
                    (<np.float64_t*>dd.data)+c*k,
                    (<np.int64_t*>ii.data)+c*k,
                    (<np.float64_t*>xx.data)+c*self.m, 
                    k, 
                    eps,
                    p, 
                    distance_upper_bound,
                    <np.float64_t*>cperiod.data)
        if single:
            if k==1:
                return dd[0,0], ii[0,0]
            else:
                return dd[0], ii[0]
        else:
            if k==1:
                return np.reshape(dd[...,0],retshape), np.reshape(ii[...,0],retshape)
            else:
                return np.reshape(dd,retshape+(k,)), np.reshape(ii,retshape+(k,))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def chainHOP_get_dens(cKDTree self, object omass, int num_neighbors=65, \
            int nMerge=6):
        """ query the tree for the nearest neighbors, to get the density
            of particles for chainHOP.
        
        Parameters:
        ===========
        
        mass: A array-like list of the masses of the particles, in the same
            order as the data that went into building the kd tree.
        
        num_neighbors: Optional, the number of neighbors to search for and to
            use in the density calculation. Default is 65, and is probably what
            one should stick with.
        
        nMerge: The number of nearest neighbor tags to return for each particle.
        
        Returns:
        ========
        
        dens: An array of the densities for each particle, in the same order
            as the input data.
        
        tags: A two-dimensional array of the indexes, nMerge nearest neighbors
            for each particle.
        
        """
        
        # We're no np.int64_ter returning all the tags in this step.
        # We do it chunked, in find_chunk_nearest_neighbors.
        #cdef np.ndarray[np.int64_t, ndim=2] tags
        cdef np.ndarray[np.float64_t, ndim=1] dens
        cdef int i, pj, j
        cdef np.float64_t ih2, fNorm, r2, rs
        
        #tags = np.empty((self.n, nMerge), dtype="int64")
        dens = np.zeros(self.n, dtype="float64")
        cdef np.ndarray[np.float64_t, ndim=2] local_data = self.data

        cdef np.ndarray[np.float64_t, ndim=1] mass = np.array(omass).astype("float64")
        cdef np.float64_t ipi = 1.0/np.pi
        
        cdef np.float64_t *query = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * self.m)
        cdef np.float64_t *dist_temp = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * num_neighbors)
        cdef np.int64_t *tags_temp = <np.int64_t *> alloca(
                    sizeof(np.int64_t) * num_neighbors)
        cdef np.float64_t period[3]
        for i in range(3): period[i] = 1.0

        for i in range(self.n):
            for j in range(self.m):
                query[j] = local_data[i,j]
            self.__query(dist_temp, tags_temp,
                         query, num_neighbors, 0.0, 
                         2, infinity, period)
            
            #calculate the density for this particle
            ih2 = -1
            for j in range(num_neighbors):
                ih2 = dmax(ih2, dist_temp[j])
            ih2 = 4.0/ih2
            fNorm = 0.5*(ih2**1.5)*ipi
            for j in range(num_neighbors):
                pj = tags_temp[j]
                r2 = dist_temp[j] * ih2
                rs = 2.0 - (r2**0.5)
                if (r2 < 1.0):
                    rs = (1.0 - 0.75*rs*r2)
                else:
                    rs = 0.25*rs*rs*rs
                rs = rs * fNorm
                dens[i] = dens[i] + rs * mass[pj]
                dens[pj] = dens[pj] + rs * mass[i]

            # store nMerge nearest neighbors
            #tags[i,:] = tags_temp[:nMerge]
        
        #return (dens, tags)
        return dens

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def find_chunk_nearest_neighbors(cKDTree self, int start, int finish, \
        int num_neighbors=65):
        """ query the tree in chunks, between start and finish, recording the
            nearest neighbors.
        
        Parameters:
        ===========
        
        start: The starting point in the dataset for this search.
        
        finish: The ending point in the dataset for this search.
        
        num_neighbors: Optional, the number of neighbors to search for.
            The default is 65.
        
        Returns:
        ========
        
        chunk_tags: A two-dimensional array of the nearest neighbor tags for the
            points in this search.
        
        """
        
        cdef np.ndarray[np.int64_t, ndim=2] chunk_tags
        cdef np.ndarray[np.float64_t, ndim=2] local_data = self.data
        cdef int i, j
        
        chunk_tags = np.empty((finish-start, num_neighbors), dtype="int64")
        cdef np.float64_t *query = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * self.m)
        cdef np.float64_t *dist_temp = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * num_neighbors)
        cdef np.int64_t *tags_temp = <np.int64_t *> alloca(
                    sizeof(np.int64_t) * num_neighbors)
        cdef np.float64_t period[3]
        for i in range(3): period[i] = 1.0

        for i in range(finish-start):
            for j in range(self.m):
                query[j] = local_data[i+start,j]
            self.__query(dist_temp, tags_temp,
                         query, num_neighbors, 0.0, 
                         2, infinity, period)
            for j in range(num_neighbors):
                chunk_tags[i,j] = tags_temp[j]
        
        return chunk_tags

    def chainHOP_preconnect(self, np.ndarray[np.int64_t, ndim=1] chainID,
                                  np.ndarray[np.float64_t, ndim=1] density,
                                  np.ndarray[np.float64_t, ndim=1] densest_in_chain,
                                  np.ndarray bis_inside,
                                  np.ndarray bsearch_again,
                                  np.float64_t peakthresh,
                                  np.float64_t saddlethresh,
                                  int nn, int nMerge,
                                  object chain_map):
        cdef np.ndarray[np.int32_t, ndim=1] is_inside
        cdef np.ndarray[np.int32_t, ndim=1] search_again
        cdef np.ndarray[np.float64_t, ndim=2] pos 
        cdef np.int64_t thisNN, thisNN_chainID, same_count
        cdef np.float64_t *query = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * self.m)
        cdef np.float64_t *dist_temp = <np.float64_t *> alloca(
                    sizeof(np.float64_t) * nn)
        cdef np.int64_t *tags_temp = <np.int64_t *> alloca(
                    sizeof(np.int64_t) * nn)
        cdef np.float64_t period[3], thisNN_max_dens, boundary_density
        cdef int i, j, npart, chainID_i, part_mas_dens
        is_inside = bis_inside.astype("int32")
        search_again = bsearch_again.astype("int32")
        pos = self.data
        npart = pos.shape[0]
        for i in range(3): period[i] = 1.0
        for i in xrange(npart):
            # Don't consider this particle if it's not part of a chain.
            if chainID[i] < 0: continue
            chainID_i = chainID[i]
            # If this particle is in the padding, don't make a connection.
            if not is_inside[i]: continue
            # Find this particle's chain max_dens.
            part_max_dens = densest_in_chain[chainID_i]
            # We're only connecting >= peakthresh chains now.
            if part_max_dens < peakthresh: continue
            # Loop over nMerge closest nearest neighbors.
            for j in range(self.m):
                query[j] = pos[i,j]
            self.__query(dist_temp, tags_temp,
                         query, nn, 0.0, 
                         2, infinity, period)
            same_count = 0
            for j in xrange(int(nMerge+1)):
                thisNN = tags_temp[j+1] # Don't consider ourselves at tags_temp[0]
                thisNN_chainID = chainID[thisNN]
                # If our neighbor is in the same chain, move on.
                # Move on if these chains are already connected:
                if chainID_i == thisNN_chainID or \
                        thisNN_chainID in chain_map[chainID_i]:
                    same_count += 1
                    continue
                # Everything immediately below is for
                # neighboring particles with a chainID. 
                if thisNN_chainID >= 0:
                    # Find thisNN's chain's max_dens.
                    thisNN_max_dens = densest_in_chain[thisNN_chainID]
                    # We're only linking peakthresh chains
                    if thisNN_max_dens < peakthresh: continue
                    # Calculate the two groups boundary density.
                    boundary_density = (density[thisNN] + density[i]) / 2.
                    # Don't connect if the boundary is too low.
                    if boundary_density < saddlethresh: continue
                    # Mark these chains as related.
                    chain_map[thisNN_chainID].add(chainID_i)
                    chain_map[chainID_i].add(thisNN_chainID)
            if same_count == nMerge + 1:
                # All our neighbors are in the same chain already, so 
                # we don't need to search again.
                search_again[i] = 0
        return search_again



import setuptools
from numpy.distutils.core import setup
from numpy.distutils.misc_util import Configuration


def configuration(parent_package='', top_path=None):

    config = Configuration('delaunay', parent_package, top_path)

    config.add_extension("_delaunay",
        sources=["_delaunay.cpp", "VoronoiDiagramGenerator.cpp",
            "delaunay_utils.cpp", "natneighbors.cpp"],
        include_dirs=['.'],
    )

    return config

if __name__ == '__main__':
    setup(**configuration(top_path='').todict())



"""
API for yt.visualization

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: J.S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: MSU
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from color_maps import \
    add_cmap

from plot_collection import \
    PlotCollection, \
    PlotCollectionInteractive, \
    concatenate_pdfs, \
    get_multi_plot

from fixed_resolution import \
    FixedResolutionBuffer, \
    ObliqueFixedResolutionBuffer

from image_writer import \
    multi_image_composite, \
    write_bitmap, \
    write_image, \
    map_to_colors, \
    splat_points, \
    annotate_image, \
    apply_colormap, \
    scale_image, \
    write_projection, \
    write_fits

from plot_modifications import \
    PlotCallback, \
    callback_registry

from easy_plots import \
    plot_type_registry

from streamlines import \
    Streamlines

from plot_window import \
    SlicePlot, \
    OffAxisSlicePlot, \
    ProjectionPlot, \
    OffAxisProjectionPlot
    




"""
This is an interface to MatPlotLib <http://matplotlib.sf.net> to plot
irregularly shaped grids, with the presumption that at any point we could have
data that is "hidden" in deeper levels of refinement.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from yt.funcs import *
from _mpl_imports import *
import _MPL
from .plot_modifications import callback_registry
from yt.utilities.definitions import \
    x_dict, \
    y_dict, \
    axis_names
from .color_maps import yt_colormaps, is_colormap
from yt.utilities.exceptions import YTNoDataInObjectError

class CallbackRegistryHandler(object):
    def __init__(self, plot):
        self.plot = plot
        self._callbacks = {}

    def __getitem__(self, item):
        if item not in self._callbacks:
            raise KeyError(item)
        cb = self._callbacks[item]

        @wraps(cb)
        def get_wrapper(*args, **kwargs):
            cbo = cb(*args, **kwargs)
            return self.plot.add_callback(cbo)

        return get_wrapper

    def __setitem__(self, item, val):
        self._callbacks[item] = val

    def __delitem__(self, item):
        del self._callbacks[item]

    def __iter__(self):
        for k in sorted(self._callbacks): yield k

    def keys(self):
        return self._callbacks.keys()

class RavenPlot(object):

    datalabel = None
    colorbar = None
    xlim = None
    ylim = None
    def __init__(self, data, fields, figure = None, axes=None, size=(10,8)):
        self.data = data
        self.fields = fields
        self.size = size
        self.set_autoscale(True)
        self.im = defaultdict(lambda: "")
        self["ParameterFile"] = "%s" % self.data.pf
        self.pf = self.data.pf
        self.axis_names = {}
        if not figure:
            self._figure = matplotlib.figure.Figure(size)
        else:
            self._figure = figure
        if not axes:
            self._axes = self._figure.add_subplot(1,1,1)
        else:
            self._axes = axes
        self._callbacks = []
        self._setup_callback_registry()

    def set_autoscale(self, val):
        self.do_autoscale = val

    def __getitem__(self, item):
        return self.data[item] # Should be returned in CGS

    def save_image(self, prefix, format="png", override=True, force_save=False,
                   figure_canvas = None):
        """
        Save this plot image.  Will generate a filename based on the *prefix*,
        *format*.  *override* will force no filename generation beyond the
        prefix.
        """
        self._redraw_image()
        if not override:
            self._generate_prefix(prefix)
            my_prefix = self.prefix
        else:
            my_prefix = prefix
        fn = ".".join([my_prefix, format])
        if figure_canvas is None:
            figure_canvas = FigureCanvasAgg
        canvas = figure_canvas(self._figure)
        if force_save:
            canvas.print_figure(fn)
        else:
            only_on_root(canvas.print_figure, fn)
        self["Type"] = self._type_name
        self["GeneratedAt"] = self.data.pf.unique_identifier
        return fn

    def open_in_webbrowser(self):
        from mplh5canvas.backend_h5canvas import FigureCanvasH5Canvas
        canv = FigureCanvasH5Canvas(self._figure)
        canv.draw()

    def save_to_pdf(self, f):
        self._redraw_image()
        canvas = FigureCanvasPdf(self._figure)
        original_figure_alpha = self._figure.patch.get_alpha()
        self._figure.patch.set_alpha(0.0)
        original_axes_alpha = []
        for ax in self._figure.axes:
            patch = ax.patch
            original_axes_alpha.append(patch.get_alpha())
            patch.set_alpha(0.0)

        canvas.print_pdf(f)

        self._figure.set_alpha(original_figure_alpha)
        for ax, alpha in zip(self._figure.axes,original_axes_alpha):
            ax.patch.set_alpha(alpha)

    def _redraw_image(self):
        pass

    def _generate_prefix(self):
        pass

    def set_xlim(self, xmin, xmax):
        """
        Set the x boundaries of this plot.
        """
        self.xlim = (xmin,xmax)

    def set_ylim(self, ymin, ymax):
        """
        Set the y boundaries of this plot.
        """
        self.ylim = (ymin,ymax)


    def set_zlim(self, zmin, zmax, dex=None, nticks=None, ticks=None, minmaxtick=False):
        """
        Set the z boundaries of this plot.

        Only ONE of the following options can be specified. If all 3 are
        specified, they will be used in the following precedence order:

        * ``ticks`` - a list of floating point numbers at which to put ticks
        * ``minmaxtick`` - display DEFAULT ticks with min & max also displayed
        * ``nticks`` - if ticks not specified, can automatically determine a
          number of ticks to be evenly spaced in log space
        """
        # This next call fixes some things, but is slower...
        self._redraw_image()
        self.set_autoscale(False)
        if (zmin in (None,'min')) or (zmax in (None,'max')):    
            imbuff = self._axes.images[-1]._A
            if zmin == 'min':
                zmin = np.nanmin(imbuff[np.nonzero(imbuff)])
                if dex is not None:
                    zmax = min(zmin*10**(dex),np.nanmax(imbuff))
            if zmax == 'max':
                zmax = np.nanmax(imbuff)
                if dex is not None:
                    zmin = max(zmax/(10**(dex)),np.nanmin(imbuff))
        if self.colorbar is not None:
            if ticks is not None:
                ticks = np.sort(ticks)
                self.colorbar.locator = matplotlib.ticker.FixedLocator(ticks)
                self.colorbar.formatter = matplotlib.ticker.FixedFormatter(["%0.2e" % (x) for x in ticks])
            elif minmaxtick:
                if self.log_field: 
                    ticks = np.array(self.colorbar._ticker()[1],dtype='float')
                    ticks = [zmin] + ticks.tolist() + [zmax]
                    self.colorbar.locator = matplotlib.ticker.FixedLocator(ticks)
                    self.colorbar.formatter = matplotlib.ticker.FixedFormatter(["%0.2e" % (x) for x in ticks])
                else:
                    mylog.error('Sorry, we do not support minmaxtick for linear fields.  It likely comes close by default')
            elif nticks is not None:
                if self.log_field:
                    lin = np.linspace(np.log10(zmin),np.log10(zmax),nticks)
                    self.colorbar.locator = matplotlib.ticker.FixedLocator(10**lin)
                    self.colorbar.formatter = matplotlib.ticker.FixedFormatter(["%0.2e" % (10**x) for x in lin])
                else: 
                    lin = np.linspace(zmin,zmax,nticks)
                    self.colorbar.locator = matplotlib.ticker.FixedLocator(lin)
                    self.colorbar.formatter = matplotlib.ticker.FixedFormatter(["%0.2e" % x for x in lin])

            else:
                if hasattr(self,'_old_locator'):
                    self.colorbar.locator = self._old_locator
                if hasattr(self,'_old_formatter'):
                    self.colorbar.formatter = self._old_formatter
        self.norm.autoscale(np.array([zmin,zmax], dtype='float64'))
        self.image.changed()
        if self.colorbar is not None:
            mpl_notify(self.image, self.colorbar)

    def set_cmap(self, cmap):
        """
        Change the colormap of this plot to *cmap*.
        """
        if isinstance(cmap, types.StringTypes):
            if str(cmap) in yt_colormaps:
                cmap = yt_colormaps[str(cmap)]
            elif hasattr(matplotlib.cm, cmap):
                cmap = getattr(matplotlib.cm, cmap)
        if not is_colormap(cmap) and cmap is not None:
            raise RuntimeError("Colormap '%s' does not exist!" % str(cmap))
        else:
            self.cmap = cmap
        
    def __setitem__(self, item, val):
        self.im[item] = val

    def add_callback(self, func):
        """
        Add *func* as a callback to this plot.  *func* will be called with this
        plot as its first argument every time the plot is redrawn.  Returns the
        id of the callback (for use with :meth:`remove_callback`.)
        """
        self._callbacks.append(func)
        return len(self._callbacks)-1

    def remove_callback(self, id):
        """
        Given an *id*, remove that index in the callbacks list.
        """
        self._callbacks[id] = lambda a: None

    def _run_callbacks(self):
        self._axes.patches = []
        self._axes.collections = []
        self._axes.texts = []
        for cb in self._callbacks:
            cb(self)

    def set_label(self, label):
        """
        Set the datalabel to *label*.  (This has different meanings based on
        the plot.)
        """
        self.datalabel = label
        if self.colorbar != None: self.colorbar.set_label(str(label))

    def setup_domain_edges(self, axis, periodic=False):
        DLE = self.data.pf.domain_left_edge
        DRE = self.data.pf.domain_right_edge
        DD = float(periodic)*(DRE - DLE)
        if axis < 3:
            xax = x_dict[axis]
            yax = y_dict[axis]
            self.xmin = DLE[xax] - DD[xax]
            self.xmax = DRE[xax] + DD[xax]
            self.ymin = DLE[yax] - DD[yax]
            self.ymax = DRE[yax] + DD[yax]
            self._period = (DD[xax], DD[yax])
            self._edges = ( (DLE[xax], DRE[xax]), (DLE[yax], DRE[yax]) )
        else:
            # Not quite sure how to deal with this, particularly
            # in the Orion case.  Cutting planes are tricky.
            self.xmin = self.ymin = 0.0
            self.xmax = self.ymax = 1.0

    def _setup_callback_registry(self):
        self.modify = CallbackRegistryHandler(self)
        for c in callback_registry.values():
            if not hasattr(c, '_type_name'): continue
            self.modify[c._type_name] = c

    def _pretty_name(self):
        width = self.im.get("Width", "NA")
        unit = self.im.get("Unit", "NA")
        field = self.axis_names.get("Z", self.axis_names.get("Field1"))
        if hasattr(self.data, "_data_source"):
            data = self.data._data_source
        else:
            data = self.data
        return "%s: %s (%s %s) %s" % (self._type_name,
            field, width, unit, data)

class VMPlot(RavenPlot):
    _antialias = True
    _period = (0.0, 0.0)
    _edges = True
    def __init__(self, data, field, figure = None, axes = None,
                 use_colorbar = True, size=None, periodic = False):
        fields = ['X', 'Y', field, 'X width', 'Y width']
        if not size:
            size = (10,8)
            if not use_colorbar: size=(8,8)
        RavenPlot.__init__(self, data, fields, figure, axes, size=size)
        self._figure.subplots_adjust(hspace=0, wspace=0, bottom=0.0,
                                    top=1.0, left=0.0, right=1.0)
        self.setup_domain_edges(self.data.axis, periodic)
        self.cmap = None
        self.label_kws = {}
        self.__setup_from_field(field)
        self.__init_temp_image(use_colorbar)

    def __setup_from_field(self, field):
        self.set_log_field(self.pf.field_info[field].take_log)
        self.axis_names["Z"] = field

    def set_log_field(self, val):
        if val:
            self.log_field = True
            self.norm = matplotlib.colors.LogNorm()
            ttype = matplotlib.ticker.LogFormatter
        else:
            self.log_field = False
            self.norm = matplotlib.colors.Normalize()
            ttype = matplotlib.ticker.ScalarFormatter
        if self.colorbar:
            self.colorbar.set_norm(self.norm)
            self.colorbar.formatter = ttype()

    def __init_temp_image(self, setup_colorbar):
        temparray = np.ones(self.size)
        self.image = \
            self._axes.imshow(temparray, interpolation='nearest',
                             norm = self.norm, aspect=1.0, picker=True,
                             origin='lower')
        self._axes.set_xticks(())
        self._axes.set_yticks(())
        self._axes.set_ylabel("")
        self._axes.set_xlabel("")
        if setup_colorbar:
            self.colorbar = self._figure.colorbar(self._axes.images[-1], \
                                                  extend='neither', \
                                                  shrink=0.95)
            self._old_locator = self.colorbar.locator
            self._old_formatter = self.colorbar.formatter
        else:
            self.colorbar = None
        self.set_width(1,'unitary')

    def _get_buff(self, width=None):
        x0, x1 = self.xlim
        y0, y1 = self.ylim
        if width is None:
            l, b, width, height = mpl_get_bounds(self._axes.bbox)
        else:
            height = width
        self.pix = (width,height)
        # 'px' == pixel x, or x in the plane of the slice
        # 'x' == actual x
        aa = int(self._antialias)
        if self._edges is True or \
           x0 < self._edges[0][0] or x1 > self._edges[0][1] or \
           y0 < self._edges[1][0] or y1 > self._edges[1][1]:
            check_period = 1
        else:
            check_period = 0
        buff = _MPL.Pixelize(self.data['px'],
                            self.data['py'],
                            self.data['pdx'],
                            self.data['pdy'],
                            self[self.axis_names["Z"]],
                            int(height), int(width),
                            (x0, x1, y0, y1),aa,self._period,
                            check_period).transpose()
        return buff

    def _redraw_image(self, *args):
        buff = self._get_buff()
        if self[self.axis_names["Z"]].size == 0:
            raise YTNoDataInObjectError(self.data)
        mylog.debug("Received buffer of min %s and max %s (data: %s %s)",
                    np.nanmin(buff), np.nanmax(buff),
                    self[self.axis_names["Z"]].min(),
                    self[self.axis_names["Z"]].max())
        if self.log_field:
            bI = np.where(buff > 0)
            if len(bI[0]) == 0:
                newmin = 1e-99
                newmax = 1e-99
            else:
                newmin = np.nanmin(buff[bI])
                newmax = np.nanmax(buff[bI])
        else:
            newmin = np.nanmin(buff)
            newmax = np.nanmax(buff)
        aspect = (self.ylim[1]-self.ylim[0])/(self.xlim[1]-self.xlim[0])
        if self.image._A.size != buff.size:
            self._axes.clear()
            self.image = \
                self._axes.imshow(buff, interpolation='nearest', norm = self.norm,
                                aspect=aspect, picker=True, origin='lower')
        else:
            self.image.set_data(buff)
        if self._axes.get_aspect() != aspect: self._axes.set_aspect(aspect)
        if self.do_autoscale:
            self.norm.autoscale(np.array((newmin,newmax), dtype='float64'))
        self._reset_image_parameters()
        self._run_callbacks()

    def _reset_image_parameters(self):
        self._axes.set_xticks(())
        self._axes.set_yticks(())
        self._axes.set_ylabel("")
        self._axes.set_xlabel("")
        if self.cmap:
            self.image.set_cmap(self.cmap)
        if self.colorbar != None:
            self.image.set_norm(self.norm)
            self.colorbar.set_norm(self.norm)
            if self.cmap: self.colorbar.set_cmap(self.cmap)
            if self.do_autoscale: mpl_notify(self.image, self.colorbar)
        self._autoset_label()

    def set_xlim(self, xmin, xmax):
        self.xlim = (xmin,xmax)

    def set_ylim(self, ymin, ymax):
        self.ylim = (ymin,ymax)

    def _generate_prefix(self, prefix):
        self.prefix = "_".join([prefix, self._type_name, \
            axis_names[self.data.axis], self.axis_names['Z']])
        self["Field1"] = self.axis_names["Z"]
        self["Field2"] = None
        self["Field3"] = None

    def set_width(self, width, unit):
        self["Unit"] = str(unit)
        self["Width"] = float(width)
        if isinstance(unit, types.StringTypes):
            unit = self.data.pf[str(unit)]
        self.width = width / unit
        self._refresh_display_width()

    def _refresh_display_width(self, width=None):
        if width:
            self.width = width
        else:
            width = self.width
        if iterable(width):
            width_x, width_y = width
        else:
            width_x = width
            width_y = width
        l_edge_x = self.data.center[x_dict[self.data.axis]] - width_x/2.0
        r_edge_x = self.data.center[x_dict[self.data.axis]] + width_x/2.0
        l_edge_y = self.data.center[y_dict[self.data.axis]] - width_y/2.0
        r_edge_y = self.data.center[y_dict[self.data.axis]] + width_y/2.0
        self.set_xlim(max(l_edge_x,self.xmin), min(r_edge_x,self.xmax))
        self.set_ylim(max(l_edge_y,self.ymin), min(r_edge_y,self.ymax))
        self._redraw_image()

    def autoscale(self):
        zmin = np.nanmin(self._axes.images[-1]._A)
        zmax = np.nanmax(self._axes.images[-1]._A)
        self.set_zlim(zmin, zmax)

    def switch_y(self, *args, **kwargs):
        pass

    def switch_x(self, *args, **kwargs):
        pass

    def switch_z(self, field):
        self.set_log_field(self.pf.field_info[field].take_log)
        self.axis_names["Z"] = field
        self._redraw_image()

    def selfSetup(self):
        pass

    def _autoset_label(self):
        if self.datalabel is None:
            field_name = self.axis_names["Z"]
            proj = "Proj" in self._type_name and \
                   self.data._weight is None
            data_label = self.pf.field_info[field_name].get_label(proj)
        else: data_label = self.datalabel
        if self.colorbar != None:
            self.colorbar.set_label(str(data_label), **self.label_kws)


class FixedResolutionPlot(VMPlot):

    # This is a great argument in favor of changing the name
    # from VMPlot to something else

    _type_name = "FixedResolution"

    def _get_buff(self, width=None):
        return self.data[self.axis_names["Z"]]

    def set_width(self, width, unit):
        #mylog.debug("Not changing FixedResolution width")
        self._refresh_display_width()

    def _refresh_display_width(self, width=None):
        self.set_xlim(self.data.bounds[0], self.data.bounds[1])
        self.set_ylim(self.data.bounds[2], self.data.bounds[3])
        self._redraw_image()

    def setup_domain_edges(self, *args, **kwargs):
        return

class PCSlicePlot(VMPlot):
    _type_name = "Slice"

    def show_velocity(self, factor = 16, bv_radius = None):
        xax = x_dict[self.data.axis]
        yax = y_dict[self.data.axis]
        xf = "%s-velocity" % (axis_names[xax])
        yf = "%s-velocity" % (axis_names[yax])
        if bv_radius is not None:
            sp = self.data.pf.h.sphere(self.data.center, bv_radius)
            bv = sp.quantities["BulkVelocity"]()
            self.data[xf + "_bv"] = self.data[xf] - bv[xax]
            self.data[yf + "_bv"] = self.data[yf] - bv[yax]
            xf += "_bv"; yf += "_bv"
        from Callbacks import QuiverCallback
        self.add_callback(QuiverCallback(xf, yf, factor))

class NNVMPlot:
    def _get_buff(self, width=None):
        import yt.utilities.delaunay as de
        x0, x1 = self.xlim
        y0, y1 = self.ylim
        if width is None:
            l, b, width, height = mpl_get_bounds(self._axes.bbox)
        else:
            height = width
        self.pix = (width,height)
        numPoints_x = int(width)
        numPoints_y = int(width)
        dx = numPoints_x / (x1-x0)
        dy = numPoints_y / (y1-y0)
        xlim = np.logical_and(self.data["px"]+2.0*self.data['pdx'] >= x0,
                              self.data["px"]-2.0*self.data['pdx'] <= x1)
        ylim = np.logical_and(self.data["py"]+2.0*self.data['pdy'] >= y0,
                              self.data["py"]-2.0*self.data['pdy'] <= y1)
        wI = np.where(np.logical_and(xlim,ylim))
        xi, yi = np.mgrid[0:numPoints_x, 0:numPoints_y]
        x = (self.data["px"][wI]-x0)*dx
        y = (self.data["py"][wI]-y0)*dy
        z = self.data[self.axis_names["Z"]][wI]
        if self.log_field: z=np.log10(z)
        buff = de.Triangulation(x,y).nn_interpolator(z)(xi,yi)
        buff = buff.clip(z.min(), z.max())
        if self.log_field: buff = 10**buff
        return buff.transpose()


class PCSlicePlotNaturalNeighbor(NNVMPlot, PCSlicePlot):
    _type_name = "NNSlice"

class PCProjectionPlot(VMPlot):

    _type_name = "Projection"

    def switch_z(self, field):
        mylog.warning("Choosing not to change the field of a projection instance")

    def _generate_prefix(self, prefix):
        VMPlot._generate_prefix(self, prefix)
        if self.data._weight is not None:
            self.prefix += "_%s" % (self.data._weight)

class PCProjectionPlotNaturalNeighbor(NNVMPlot, PCProjectionPlot):
    _type_name = "NNProj"

class CuttingPlanePlot(PCSlicePlot):

    _type_name = "CuttingPlane"
    def _get_buff(self, width=None):
        px_min, px_max = self.xlim
        py_min, py_max = self.ylim
        if width is None:
            l, b, width, height = mpl_get_bounds(self._axes.bbox)
        else:
            height = width
        self.pix = (width,height)
        indices = np.argsort(self.data['dx'])[::-1]
        buff = _MPL.CPixelize( self.data['x'], self.data['y'], self.data['z'],
                               self.data['px'], self.data['py'],
                               self.data['pdx'], self.data['pdy'], self.data['pdz'],
                               self.data.center, self.data._inv_mat, indices,
                               self.data[self.axis_names['Z']],
                               int(width), int(width),
                               (px_min, px_max, py_min, py_max)).transpose()
        return buff

    def _refresh_display_width(self, width=None):
        
        if width:
            self.width = width
        else:
            width = self.width
        if iterable(width):
            width_x, width_y = width
        else:
            width_x = width
            width_y = width
        l_edge_x = -width_x/2.0
        r_edge_x = +width_x/2.0
        l_edge_y = -width_y/2.0
        r_edge_y = +width_y/2.0
        self.set_xlim(l_edge_x, r_edge_x) # We have no real limits
        self.set_ylim(l_edge_y, r_edge_y) # At some point, perhaps calculate them?
        self._redraw_image()

class ParticlePlot(RavenPlot):

    _type_name = "ParticlePlot"
    def __init__(self, data, axis, width, p_size=1.0, col='k', stride=1.0,
                 figure = None, axes = None):
        kwargs = {}
        if figure is None: kwargs['size'] = (8,8)
        RavenPlot.__init__(self, data, [], figure, axes, **kwargs)
        self._figure.subplots_adjust(hspace=0, wspace=0, bottom=0.0,
                                    top=1.0, left=0.0, right=1.0)
        self.axis = axis
        self.setup_domain_edges(axis)
        self.axis_names['Z'] = col
        self.modify["particles"](width, p_size, col, stride)
        self.set_width(1,'unitary')

    def _redraw_image(self, *args):
        self._axes.clear() # To help out the colorbar
        self._reset_image_parameters()
        self._run_callbacks()

    def _generate_prefix(self, prefix):
        self.prefix = "_".join([prefix, self._type_name, \
            axis_names[self.axis], self.axis_names['Z']])
        self["Field1"] = self.axis_names["Z"]
        self["Field2"] = None
        self["Field3"] = None

    def set_width(self, width, unit):
        self["Unit"] = str(unit)
        self["Width"] = float(width)
        if isinstance(unit, types.StringTypes):
            unit = self.data.pf[str(unit)]
        self.width = width / unit
        self._refresh_display_width()

    def _refresh_display_width(self, width=None):
        if width:
            self.width = width
        else:
            width = self.width
        if iterable(width):
            width_x, width_y = width
        else:
            width_x = width
            width_y = width
        l_edge_x = self.data.center[x_dict[self.axis]] - width_x/2.0
        r_edge_x = self.data.center[x_dict[self.axis]] + width_x/2.0
        l_edge_y = self.data.center[y_dict[self.axis]] - width_y/2.0
        r_edge_y = self.data.center[y_dict[self.axis]] + width_y/2.0
        self.set_xlim(max(l_edge_x,self.xmin), min(r_edge_x,self.xmax))
        self.set_ylim(max(l_edge_y,self.ymin), min(r_edge_y,self.ymax))
        self._redraw_image()

    def _reset_image_parameters(self):
        self._axes.set_xticks(())
        self._axes.set_yticks(())
        self._axes.set_ylabel("")
        self._axes.set_xlabel("")
        l, b, width, height = mpl_get_bounds(self._axes.bbox)
        self._axes.set_xlim(0, width)
        self._axes.set_ylim(0, height)

class ProfilePlot(RavenPlot):
    _x_label = None
    _y_label = None

    def setup_bins(self, field, func=None):
        if field in self.pf.field_info and self.pf.field_info[field].take_log:
            log_field = True
            if func: func('log')
        else:
            log_field = False
            if func: func('linear')
        mylog.debug("Field: %s, log_field: %s", field, log_field)
        return log_field

    def set_x_label(self, label):
        self._axes.set_xlabel(label)
        self._x_label = label

    def set_y_label(self, label):
        self._axes.set_ylabel(label)
        self._y_label = label

    def _autoset_label(self, field, function, axis):
        label = getattr(self, '_%s_label' % axis)
        if label is None: label = self.pf.field_info[field].get_label()
        function(label)

class Profile1DPlot(ProfilePlot):

    _type_name = "Profile1D"
    def __init__(self, data, fields, id, ticker=None, 
                 figure=None, axes=None, plot_options=None):
        self._semi_unique_id = id
        RavenPlot.__init__(self, data, fields, figure, axes)

        self.axis_names["X"] = fields[0]
        self.axis_names["Y"] = fields[1]

        if plot_options is None: plot_options = {}
        self.plot_options = plot_options

        self._log_x = self.data._x_log
        self._log_y = self.setup_bins(self.fields[1])

    def _generate_prefix(self, prefix):
        self.prefix = "_".join([prefix, self._type_name,
                       str(self._semi_unique_id),
                       self.axis_names['X'], self.axis_names['Y']])
        self["Field1"] = self.axis_names["X"]
        self["Field2"] = self.axis_names["Y"]

    def _redraw_image(self):
        self._axes.clear()
        if not self._log_x and not self._log_y:
            func = self._axes.plot
        elif self._log_x and not self._log_y:
            func = self._axes.semilogx
        elif not self._log_x and self._log_y:
            func = self._axes.semilogy
        elif self._log_x and self._log_y:
            func = self._axes.loglog
        indices = np.argsort(self.data[self.fields[0]])
        func(self.data[self.fields[0]][indices],
             self.data[self.fields[1]][indices],
             **self.plot_options)
        self._autoset_label(self.fields[0], self.set_x_label, 'x')
        self._autoset_label(self.fields[1], self.set_y_label, 'y')
        if self.xlim is not None: self._axes.set_xlim(*self.xlim)
        if self.ylim is not None: self._axes.set_ylim(*self.ylim)
        self._run_callbacks()

    def set_log_field(self, val):
        if val:
            self._log_y = True
        else:
            self._log_y = False

    def switch_x(self, field, weight="CellMassMsun", accumulation=False):
        self.fields[0] = field
        self.axis_names["X"] = field
        if field not in self.data.keys():
            self.data.add_fields(field, weight, accumulation)
        self._log_x = self.setup_bins(self.fields[0])
    
    def switch_z(self, field, weight="CellMassMsun", accumulation=False):
        self.fields[1] = field
        self.axis_names["Y"] = field
        if field not in self.data.keys():
            self.data.add_fields(field, weight, accumulation)
        self._log_y = self.setup_bins(self.fields[1])
    switch_y = switch_z # Compatibility...

class PhasePlot(ProfilePlot):

    _type_name = "Profile2D"
    _xlim = None
    _ylim = None
    _z_label = None

    def __init__(self, data, fields, id, ticker=None, cmap=None,
                 figure=None, axes=None):
        self._semi_unique_id = id
        RavenPlot.__init__(self, data, fields, figure, axes)
        self.ticker = ticker
        self.image = None
        self.set_cmap(cmap)
        self._zlim = None
     
        self.axis_names["X"] = fields[0]
        self.axis_names["Y"] = fields[1]
        self.axis_names["Z"] = fields[2]

        self.x_bins = self.data[self.fields[0]]
        self.y_bins = self.data[self.fields[1]]
        self._log_x = self.data._x_log
        self._log_y = self.data._y_log
        self._log_z = self.setup_bins(self.fields[2])
        self.__init_colorbar()

    def _run_callbacks(self):
        # We sublcass to avoid the problem of replacing self.image,
        # which is a collection
        self._axes.patches = []
        self._axes.texts = []
        for cb in self._callbacks:
            cb(self)

    def __init_colorbar(self):
        temparray = np.ones((self.x_bins.size, self.y_bins.size))
        self.norm = matplotlib.colors.Normalize()
        self.image = self._axes.pcolormesh(self.x_bins, self.y_bins,
                                      temparray, shading='flat',
                                      norm=self.norm, cmap=self.cmap,
                                      rasterized=True)
        self.colorbar = self._figure.colorbar(self.image,
                                    extend='neither', shrink=0.95,
                                    format="%0.2e" )

    def set_cmap(self, cmap):
        RavenPlot.set_cmap(self, cmap)
        if self.image != None and self.cmap != None:
            self.image.set_cmap(self.cmap)

    def switch_z(self, field, weight="CellMassMsun", accumulation=False, fractional=False):
        self.fields[2] = field
        self.axis_names["Z"] = field
        if field not in self.data.keys(): self.data.add_fields(field, weight, accumulation, fractional=fractional)
        self._log_z = self.setup_bins(self.fields[2])

    def set_xlim(self, xmin, xmax):
        self._xlim = (xmin,xmax)

    def set_ylim(self, ymin, ymax):
        self._ylim = (ymin,ymax)

    def set_zlim(self, zmin, zmax, dex=None):
        """
        Set the z boundaries of this plot.
        """
        # This next call fixes some things, but is slower...
        #self._redraw_image()
        if (zmin is None) or (zmax is None):    
            if zmin == 'min':
                zmin = np.nanmin(self._axes.images[-1]._A)
                if dex is not None:
                    zmax = min(zmin*10**(dex),np.nanmax(self._axes.images[-1]._A))
            if zmax == 'max':
                zmax = np.nanmax(self._axes.images[-1]._A)
                if dex is not None:
                    zmin = max(zmax/(10**(dex)),np.nanmin(self._axes.images[-1]._A))
        self._zlim = (zmin, zmax)

    def set_log_field(self, val):
        if val:
            self._log_z = True
            self.norm = matplotlib.colors.LogNorm()
            ttype = matplotlib.ticker.LogFormatter
        else:
            self._log_z = False
            self.norm = matplotlib.colors.Normalize()
            ttype = matplotlib.ticker.ScalarFormatter
        if self.colorbar:
            self.colorbar.set_norm(self.norm)
            self.colorbar.formatter = ttype()

    def _redraw_image(self):
        vals = self.data[self.fields[2]].transpose()
        used_bin = self.data["UsedBins"].transpose()
        vmin = np.nanmin(vals[used_bin])
        vmax = np.nanmax(vals[used_bin])
        if self._zlim is not None: vmin, vmax = self._zlim
        if self._log_z:
            # We want smallest non-zero vmin
            self.norm=matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax,
                                                clip=False)
            self.ticker = matplotlib.ticker.LogLocator()
            if self._zlim is None:
                vI = np.where(vals > 0)
                vmin = vals[vI].min()
                vmax = vals[vI].max()
            self.norm.autoscale(np.array((vmin,vmax), dtype='float64'))
        else:
            self.norm=matplotlib.colors.Normalize(vmin=vmin, vmax=vmax,
                                                  clip=False)
            self.ticker = matplotlib.ticker.MaxNLocator()
        self.colorbar.set_norm(self.norm)
        if self.cmap == None:
            self.cmap = matplotlib.cm.get_cmap()
        self.cmap.set_bad("w")
        self.cmap.set_under("w")
        self.cmap.set_over("w")
        self._axes.clear()
        self._axes.set_xscale("linear")
        self._axes.set_yscale("linear")
        self.image = self._axes.pcolormesh(self.x_bins, self.y_bins, \
                                      vals, shading='flat', \
                                      norm=self.norm, cmap=self.cmap,
                                      rasterized=True)
        self._axes.set_xscale({0:"linear",1:"log"}[int(self._log_x)])
        self._axes.set_yscale({0:"linear",1:"log"}[int(self._log_y)])
        if self._xlim is not None: self._axes.set_xlim(*self._xlim)
        if self._ylim is not None: self._axes.set_ylim(*self._ylim)
        self.vals = vals

        mpl_notify(self.image, self.colorbar)
        self._autoset_label(self.fields[0], self.set_x_label, 'x')
        self._autoset_label(self.fields[1], self.set_y_label, 'y')
        self._autoset_label(self.fields[2], self.set_z_label, 'z')
        self._run_callbacks()

    def _generate_prefix(self, prefix):
        self.prefix = "_".join([str(prefix), self._type_name,
            str(self._semi_unique_id),
            self.axis_names['X'], self.axis_names['Y'],
            self.axis_names['Z'], ])
        self["Field1"] = self.axis_names["X"]
        self["Field2"] = self.axis_names["Y"]
        self["Field3"] = self.axis_names["Z"]

    def set_width(self, width, unit):
        mylog.warning("Choosing not to change the width of a phase plot instance")

    def set_z_label(self, label):
        self.colorbar.set_label(label)
        self._z_label = label

class LineQueryPlot(RavenPlot):
    _type_name = "LineQueryPlot"

    def __init__(self, data, fields, id, ticker=None,
                 figure=None, axes=None, plot_options=None):
        self._semi_unique_id = id
        RavenPlot.__init__(self, data, fields, figure, axes)

        self.axis_names["X"] = fields[0]
        self.axis_names["Y"] = fields[1]

        self._log_x = False
        if fields[1] in self.pf.field_info and \
            self.pf.field_info[fields[1]].take_log:
            self._log_y = True
        else:
            self._log_y = False

        if plot_options is None: plot_options = {}
        self.plot_options = plot_options

    def _generate_prefix(self, prefix):
        self.prefix = "_".join([prefix, self._type_name,
                       str(self._semi_unique_id),
                       self.axis_names['X'], self.axis_names['Y']])
        self["Field1"] = self.axis_names["X"]
        self["Field2"] = self.axis_names["Y"]

    def _redraw_image(self):
        self._axes.clear()
        if not self._log_x and not self._log_y:
            func = self._axes.plot
        elif self._log_x and not self._log_y:
            func = self._axes.semilogx
        elif not self._log_x and self._log_y:
            func = self._axes.semilogy
        elif self._log_x and self._log_y:
            func = self._axes.loglog
        indices = np.argsort(self.data[self.fields[0]])
        func(self.data[self.fields[0]][indices],
             self.data[self.fields[1]][indices],
             **self.plot_options)
        self._autoset_label(self.fields[0], self._axes.set_xlabel)
        self._autoset_label(self.fields[1], self._axes.set_ylabel)
        self._run_callbacks()

    def set_log_field(self, val):
        if val:
            self._log_y = True
        else:
            self._log_y = False

    def switch_x(self, field):
        self.fields[0] = field
        self.axis_names["X"] = field
    
    def switch_z(self, field):
        self.fields[1] = field
        self.axis_names["Y"] = field

    def _autoset_label(self, field_name, func):
        if self.datalabel != None:
            func(str(self.datalabel))
            return
        data_label = r"$\rm{%s}" % field_name.replace("_"," ")
        if field_name in self.pf.field_info:
            data_label += r"\/\/ (%s)" % (self.pf.field_info[field_name].get_units())
        data_label += r"$"
        func(str(data_label))


    switch_y = switch_z # Compatibility...

class ScatterPlot(LineQueryPlot):
    _type_name = "ScatterPlot"

    def _redraw_image(self):
        self._axes.clear()
        self._axes.scatter(
             self.data[self.fields[0]], self.data[self.fields[1]],
             **self.plot_options)
        xscale = {True: "log", False: "linear"}.get(self._log_x)
        yscale = {True: "log", False: "linear"}.get(self._log_y)
        self._axes.set_xscale(xscale)
        self._axes.set_yscale(yscale)
        self._autoset_label(self.fields[0], self._axes.set_xlabel)
        self._autoset_label(self.fields[1], self._axes.set_ylabel)
        self._run_callbacks()

    # We override because we don't want to ditch our collections
    def _run_callbacks(self):
        self._axes.patches = []
        self._axes.texts = []
        for cb in self._callbacks:
            cb(self)




"""
Hydro tests

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import matplotlib
import pylab
from yt.mods import *
from output_tests import SingleOutputTest, YTStaticOutputTest, create_test

class TestProjection(YTStaticOutputTest):

    field = None
    axis = None
    weight_field = None

    def run(self):
        # First we get our flattened projection -- this is the
        # Density, px, py, pdx, and pdy
        proj = self.pf.h.proj(self.axis, self.field,
                              weight_field=self.weight_field)
        # Now let's stick it in a buffer
        pixelized_proj = self.pixelize(proj, self.field)
        # We just want the values, so this can be stored
        # independently of the parameter file.
        # The .field_data attributes strip out everything other than the actual array
        # values.
        self.result = (proj.field_data, pixelized_proj.data)

    def compare(self, old_result):
        proj, pixelized_proj = self.result
        oproj, opixelized_proj = old_result

        self.compare_data_arrays(proj, oproj)
        self.compare_array_delta(
            pixelized_proj[self.field],
            opixelized_proj[self.field],
            1e-7)

    def plot(self):
        pylab.clf()
        pylab.imshow(self.result[1][self.field],
            interpolation='nearest', origin='lower')
        fn = "%s_%s_%s_projection.png" % (self.pf, self.field,
                                          self.weight_field)
        pylab.savefig(fn)
        return [fn]

class TestOffAxisProjection(YTStaticOutputTest):

    field = None
    weight_field = None

    def run(self):
        # Here proj will just be the data array.
        proj = off_axis_projection(self.pf, 
                                   (0.5 * (self.pf.domain_left_edge + 
                                           self.pf.domain_right_edge)),
                                   [1., 1., 1.], 1., 400,
                                   self.field, weight=self.weight_field)

        # values.
        self.result = proj

    def compare(self, old_result):
        proj  = self.result
        oproj = old_result

        self.compare_array_delta(proj, oproj, 1e-7)

    def plot(self):
        fn = "%s_%s_%s_off-axis_projection.png" % \
            (self.pf, self.field, self.weight_field)
        write_image(self.result, fn)
        return [fn]

class TestRay(YTStaticOutputTest):

    field = None

    def run(self):
        np.random.seed(4333)
        start_point = np.random.random(self.pf.dimensionality) * \
            (self.pf.domain_right_edge - self.pf.domain_left_edge) + \
            self.pf.domain_left_edge
        end_point   = np.random.random(self.pf.dimensionality) * \
            (self.pf.domain_right_edge - self.pf.domain_left_edge) + \
            self.pf.domain_left_edge

        # Here proj will just be the data array.
        ray = self.pf.h.ray(start_point, end_point, field=self.field)

        # values.
        self.result = ray[self.field]

    def compare(self, old_result):
        ray  = self.result
        oray = old_result

        self.compare_array_delta(ray, oray, 1e-7)

    def plot(self):
        return

class TestSlice(YTStaticOutputTest):

    field = None
    axis = None

    def run(self):
        # Here proj will just be the data array.
        slice = self.pf.h.slice(self.axis, 
                                (0.5 * (self.pf.domain_left_edge + 
                                        self.pf.domain_right_edge))[self.axis],
                                fields=self.field)
        # values.
        self.result = slice.field_data

    def compare(self, old_result):
        slice  = self.result
        oslice = old_result

        self.compare_data_arrays(slice, oslice)

    def plot(self):
        fn = "%s_%s_slice.png" % (self.pf, self.field)
        write_image(self.result[self.field], fn)
        return [fn]

# Now we create all our tests.  We are using the create_test
# function, which is a relatively simple function that takes the base class,
# a name, and any parameters that the test requires.
for axis in range(3):
    for field in ["Density", "Temperature"]:
        create_test(TestProjection, "projection_test_%s_%s" % (axis, field),
                    field = field, axis = axis)

class TestGasDistribution(YTStaticOutputTest):
    field_x = None
    field_y = None
    weight = "CellMassMsun"
    n_bins = 32

    def run(self):
        # We're NOT going to use the low-level profiling API here,
        # because we are avoiding the calculations of min/max,
        # as those should be tested in another test.
        pc = PlotCollection(self.pf, center=self.sim_center)
        p = pc.add_profile_object(self.entire_simulation,
            [self.field_x, self.field_y], x_bins = self.n_bins,
            weight=self.weight)
        # The arrays are all stored in a dictionary hanging off the profile
        # object
        self.result = p.data.field_data
                    
    def compare(self, old_result):
        self.compare_data_arrays(
            self.result, old_result)

    def plot(self):
        return []

# Now we create all our tests, but we're only going to check the binning
# against Density for now.
for field in ["Temperature", "x-velocity"]:
    create_test(TestGasDistribution, "profile_density_test_%s" % field,
                field_x = "Density", field_y = field)

class Test2DGasDistribution(TestGasDistribution):
    x_bins = 128
    y_bins = 128
    field_z = "CellMassMsun"
    weight = None
    def run(self):
        # We're NOT going to use the low-level profiling API here,
        # because we are avoiding the calculations of min/max,
        # as those should be tested in another test.
        pc = PlotCollection(self.pf, center=self.sim_center)
        p = pc.add_phase_object(self.entire_simulation,
            [self.field_x, self.field_y, self.field_z], x_bins = self.x_bins, y_bins = self.y_bins,
            weight=self.weight)
        # The arrays are all stored in a dictionary hanging off the profile
        # object
        self.result = p.data.field_data




"""
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Major library imports
from vm_panner import VariableMeshPanner
from numpy import linspace, meshgrid, pi, sin, mgrid, zeros

# Enthought library imports
from enthought.enable.api import Component, ComponentEditor, Window
from enthought.traits.api import HasTraits, Instance, Button, Any, Callable, \
        on_trait_change, Bool, DelegatesTo, List, Enum, Int, Property, Str, \
        cached_property
from enthought.traits.ui.api import \
        Group, View, Item, VGroup, InstanceEditor

# Chaco imports
from enthought.chaco.api import ArrayPlotData, jet, Plot, HPlotContainer, \
        ColorBar, DataRange1D, DataRange2D, LinearMapper, ImageData, \
        CMapImagePlot, OverlayPlotContainer
from enthought.chaco.tools.api import PanTool, ZoomTool, RangeSelection, \
        RangeSelectionOverlay, RangeSelection
from enthought.chaco.tools.image_inspector_tool import ImageInspectorTool, \
     ImageInspectorOverlay

if not hasattr(DataRange2D, "_subranges_updated"):
    print "You'll need to add _subranges updated to enthought/chaco/data_range_2d.py"
    print 'Add this at the correct indentation level:'
    print
    print '    @on_trait_change("_xrange.updated,_yrange.updated")'
    print '    def _subranges_updated(self):'
    print '        self.updated = True'
    print
    raise RuntimeError

# We like the algae colormap; for now we re-implement it here.

from enthought.chaco.api import \
    ColorMapper, \
    color_map_functions, color_map_dict, color_map_name_dict

def algae(range, **traits):
    _data = {'red':   ((0.0, 80/256., 80/256.),
                       (0.2, 0.0, 0.0),
                       (0.4, 0.0, 0.0),
                       (0.6, 256/256., 256/256.),
                       (0.95, 256/256., 256/256.),
                       (1.0, 150/256., 150/256.)),
             'green': ((0.0, 0/256., 0/256.),
                       (0.2, 0/256., 0/256.),
                       (0.4, 130/256., 130/256.),
                       (0.6, 256/256., 256/256.),
                       (1.0, 0.0, 0.0)),
             'blue':  ((0.0, 80/256., 80/256.),
                       (0.2, 220/256., 220/256.),
                       (0.4, 0.0, 0.0),
                       (0.6, 20/256., 20/256.),
                       (1.0, 0.0, 0.0))}
    return ColorMapper.from_segment_map(_data, range=range, **traits)
color_map_functions.append(algae)
color_map_dict[algae] = "algae"
color_map_name_dict["algae"] = algae

class FunctionImageData(ImageData):
    # The function to call with the low and high values of the range.
    # It should return an array of values.
    func = Callable

    # A reference to a datarange
    data_range = Instance(DataRange2D)

    def __init__(self, **kw):
        # Explicitly call the AbstractDataSource constructor because
        # the ArrayDataSource ctor wants a data array
        ImageData.__init__(self, **kw)
        self.recalculate()

    @on_trait_change('data_range.updated')
    def recalculate(self):
        if self.func is not None and self.data_range is not None:
            newarray = self.func(self.data_range.low, self.data_range.high)
            ImageData.set_data(self, newarray)
        else:
            self._data = zeros((512,512),dtype=float)

    def set_data(self, *args, **kw):
        raise RuntimeError("Cannot set numerical data on a FunctionDataSource")

    def set_mask(self, mask):
        raise NotImplementedError

    def remove_mask(self):
        raise NotImplementedError

class ImagePixelizerHelper(object):
    index = None
    def __init__(self, panner, run_callbacks = False):
        self.panner = panner
        self.run_callbacks = run_callbacks

    def __call__(self, low, high):
        b = self.panner.set_low_high(low, high)
        if self.run_callbacks:
            self.panner._run_callbacks()
        if self.index is not None:
            num_x_ticks = b.shape[0] + 1
            num_y_ticks = b.shape[1] + 1
            xs = mgrid[low[0]:high[0]:num_x_ticks*1j]
            ys = mgrid[low[1]:high[1]:num_y_ticks*1j]
            self.index.set_data( xs, ys )
        return b

class ZoomedPlotUpdater(object):
    fid = None
    def __init__(self, panner, zoom_factor=4):
        """
        Supply this an a viewport_callback argument to a panner if you want to
        update a second panner in a smaller portion at higher resolution.  If
        you then set the *fid* property, you can also have it update a
        FunctionImageData datarange.  *panner* is the panner to update (not the
        one this is a callback to) and *zoom_factor* is how much to zoom in by.
        """
        self.panner = panner
        self.zoom_factor = zoom_factor

    def __call__(self, xlim, ylim):
        self.panner.xlim = xlim
        self.panner.ylim = ylim
        self.panner.zoom(self.zoom_factor)
        nxlim = self.panner.xlim
        nylim = self.panner.ylim
        if self.fid is not None:
            self.fid.data_range.set_bounds(
                (nxlim[0], nylim[0]), (nxlim[1], nylim[1]))

class VMImagePlot(HasTraits):
    plot = Instance(Plot)
    fid = Instance(FunctionImageData)
    img_plot = Instance(CMapImagePlot)
    panner = Instance(VariableMeshPanner)
    helper = Instance(ImagePixelizerHelper)
    fields = List

    def __init__(self, *args, **kwargs):
        super(VMImagePlot, self).__init__(**kwargs)
        self.add_trait("field", Enum(*self.fields))
        self.field = self.panner.field

    def _plot_default(self):
        pd = ArrayPlotData()
        plot = Plot(pd, padding = 0)
        self.fid._data = self.panner.buffer

        pd.set_data("imagedata", self.fid)

        img_plot = plot.img_plot("imagedata", colormap=algae,
                                 interpolation='nearest',
                                 xbounds=(0.0, 1.0),
                                 ybounds=(0.0, 1.0))[0]
        self.fid.data_range = plot.range2d
        self.helper.index = img_plot.index
        self.img_plot = img_plot
        return plot

    def _field_changed(self, old, new):
        self.panner.field = new
        self.fid.recalculate()

    def _fid_default(self):
        return FunctionImageData(func = self.helper)

    def _helper_default(self):
        return ImagePixelizerHelper(self.panner)

    def _panner_changed(self, old, new):
        index = self.helper.index
        self.helper = ImagePixelizerHelper(new)
        self.helper.index = index
        self.fid.func = self.helper
        self.fid.recalculate()

    def _fields_default(self):
        keys = []
        for field in self.panner.source.keys():
            if field not in ['px','py','pdx','pdy',
                             'pz','pdz','weight_field']:
                keys.append(field)
        return keys

class VariableMeshPannerView(HasTraits):

    plot = Instance(Plot)
    spawn_zoom = Button
    vm_plot = Instance(VMImagePlot)
    use_tools = Bool(True)
    full_container = Instance(HPlotContainer)
    container = Instance(OverlayPlotContainer)
    
    traits_view = View(
                    Group(
                        Item('full_container',
                             editor=ComponentEditor(size=(512,512)), 
                             show_label=False),
                        Item('field', show_label=False),
                        orientation = "vertical"),
                    width = 800, height=800,
                    resizable=True, title="Pan and Scan",
                    )

    def _vm_plot_default(self):
        return VMImagePlot(panner=self.panner)
    
    def __init__(self, **kwargs):
        super(VariableMeshPannerView, self).__init__(**kwargs)
        # Create the plot
        self.add_trait("field", DelegatesTo("vm_plot"))

        plot = self.vm_plot.plot
        img_plot = self.vm_plot.img_plot

        if self.use_tools:
            plot.tools.append(PanTool(img_plot))
            zoom = ZoomTool(component=img_plot, tool_mode="box", always_on=False)
            plot.overlays.append(zoom)
            imgtool = ImageInspectorTool(img_plot)
            img_plot.tools.append(imgtool)
            overlay = ImageInspectorOverlay(component=img_plot, image_inspector=imgtool,
                                            bgcolor="white", border_visible=True)
            img_plot.overlays.append(overlay)


        image_value_range = DataRange1D(self.vm_plot.fid)
        cbar_index_mapper = LinearMapper(range=image_value_range)
        self.colorbar = ColorBar(index_mapper=cbar_index_mapper,
                                 plot=img_plot,
                                 padding_right=40,
                                 resizable='v',
                                 width=30)

        self.colorbar.tools.append(
            PanTool(self.colorbar, constrain_direction="y", constrain=True))
        zoom_overlay = ZoomTool(self.colorbar, axis="index", tool_mode="range",
                                always_on=True, drag_button="right")
        self.colorbar.overlays.append(zoom_overlay)

        # create a range selection for the colorbar
        range_selection = RangeSelection(component=self.colorbar)
        self.colorbar.tools.append(range_selection)
        self.colorbar.overlays.append(
                RangeSelectionOverlay(component=self.colorbar,
                                      border_color="white",
                                      alpha=0.8, fill_color="lightgray"))

        # we also want to the range selection to inform the cmap plot of
        # the selection, so set that up as well
        range_selection.listeners.append(img_plot)

        self.full_container = HPlotContainer(padding=30)
        self.container = OverlayPlotContainer(padding=0)
        self.full_container.add(self.colorbar)
        self.full_container.add(self.container)
        self.container.add(self.vm_plot.plot)

class OutputSelector(HasTraits):
    outputs = List
    main_plot = Instance(VariableMeshPannerView)
    main_panner = Property(depends_on = "pf")
    weight_field = Str("Density")

    pf = Any
    source = Any
    axis = Int(0)

    traits_view = View(VGroup(
                        Item('output'),
                        Item('main_plot'),
                        )
                      )

    def __init__(self, **kwargs):
        super(OutputSelector, self).__init__(**kwargs)
        self.add_trait("output", Enum(*self.outputs))
        self.output = self.outputs[-1]
        self.main_plot

    def _output_default(self):
        return self.outputs[0]

    def _output_changed(self, old, new):
        # We get a string here
        import yt.mods
        self.pf = yt.mods.load(new, data_style="enzo_packed_3d")
        self.source = yt.mods.projload(self.pf, self.axis, "Density")
        self.main_panner.field = self.main_plot.vm_plot.field
        self.main_plot.panner = self.main_plot.vm_plot.panner = \
            self.main_plot.vm_plot.helper.panner = self.main_panner
        self.main_plot.vm_plot.field = self.main_panner.field

    def _main_plot_default(self):
        vmpv = VariableMeshPannerView(panner = self.main_panner)
        vmpv.vm_plot.helper.run_callbacks = True
        return vmpv

    @cached_property
    def _get_main_panner(self):
        return self.pf.h.image_panner(self.source, (512, 512), "Density")

def pan_and_scan_directory(dir_name):
    import glob, os
    fns = [ fn[:-10] for fn in
            glob.glob(os.path.join(dir_name, "**", "*.hierarchy")) ]
    selector = OutputSelector(outputs = fns)
    return selector



from yt.config import ytcfg
ytcfg["yt","loglevel"] = '50'
ytcfg["yt","suppressStreamLogging"] = 'True'

import hydro_tests # Just importing will register the tests!
import halo_tests
import particle_tests
from runner import RegressionTestRunner

first_runner = RegressionTestRunner("first")
first_runner.run_all_tests()
second_runner = RegressionTestRunner("second", "first")
second_runner.run_all_tests()



from yt.config import ytcfg;ytcfg["yt","__withinreason"]="True"
import os
import Pyro4
import uuid

from yt.mods import *
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    _get_comm
from yt.gui.reason.pyro_queue import \
    PyroQueueRoot, \
    PyroQueueNonRoot

comm = _get_comm(())
my_rank = comm.comm.rank

if my_rank == 0:
    my_q = PyroQueueRoot(comm)
    Pyro4.config.HMAC_KEY = uuid.uuid4().hex
    key_file = 'reason.key'
    fd = os.open(key_file, os.O_CREAT, 0600)
    os.close(fd)
    out_file = file(key_file, 'w')
    out_file.write("HMAC KEY: %s\n" % Pyro4.config.HMAC_KEY)
    out_file.close()
    mylog.info('See %s for HMAC key.', key_file)
    Pyro4.Daemon.serveSimple(
        {my_q: "yt.executor"},
        ns=False, verbose=True)
else:
    my_q = PyroQueueNonRoot(comm)
    my_q.run()



"""
Wrapping code for Oliver Hahn's RamsesRead++

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Author: Oliver Hahn <ohahn@stanford.edu>
Affiliation: KIPAC / Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Cython wrapping code for Oliver Hahn's RAMSES reader
from cython.operator cimport dereference as deref, preincrement as inc
from libc.stdlib cimport malloc, free, abs, calloc, labs

import numpy as np
cimport numpy as np
cimport cython

cdef extern from "math.h":
    double ceil(double x)
    double log2(double x)

cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1):
    if i0 > i1: return i0
    return i1

cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1):
    if i0 < i1: return i0
    return i1

cdef extern from "<vector>" namespace "std":
    cdef cppclass vector[T]:
        cppclass iterator:
            T operator*()
            iterator operator++()
            bint operator==(iterator)
            bint operator!=(iterator)
        vector()
        void push_back(T&)
        T& operator[](int)
        T& at(int)
        iterator begin()
        iterator end()

cdef extern from "<map>" namespace "std":
    cdef cppclass map[A,B]:
        pass

cdef extern from "string" namespace "std":
    cdef cppclass string:
        string(char *cstr)
        char *c_str()
        string operator*()

cdef extern from "RAMSES_typedefs.h":
    pass

cdef extern from "RAMSES_info.hh" namespace "RAMSES":
    enum codeversion:
        version1
        version2
        version3

    cdef cppclass snapshot:
        string m_filename
        codeversion m_version
        cppclass info_data:
            unsigned ncpu
            unsigned ndim
            unsigned levelmin
            unsigned levelmax
            unsigned ngridmax
            unsigned nstep_coarse
            double boxlen
            double time
            double aexp
            double H0
            double omega_m
            double omega_l
            double omega_k
            double omega_b
            double unit_l
            double unit_d
            double unit_t

        info_data m_header
        vector[double] ind_min
        vector[double] ind_max

        snapshot(string info_filename, codeversion ver)

        unsigned get_snapshot_num()
        unsigned getdomain_bykey( double key )

    #void hilbert3d(vector[double] &x, vector[double] &y, vector[double] &z,
    #               vector[double] &order, unsigned bit_length)

cdef extern from "RAMSES_amr_data.hh" namespace "RAMSES::AMR":
    cdef cppclass vec[real_t]:
        real_t x, y, z
        vec( real_t x_, real_t y_, real_t z_)
        vec ( vec& v)
        vec ( )

    # This class definition is out of date.  I have unrolled the template
    # below.
    cdef cppclass cell_locally_essential[id_t, real_t]:
        id_t m_neighbor[6]
        id_t m_father
        id_t m_son[8]
        real_t m_xg[3]
        id_t m_cpu

        char m_pos

        cell_locally_essential()

        bint is_refined(int ison)

    cdef cppclass RAMSES_cell:
        unsigned m_neighbour[6]
        unsigned m_father
        unsigned m_son[8]
        float m_xg[3]
        unsigned m_cpu

        char m_pos

        RAMSES_cell()

        bint is_refined(int ison)

    cdef cppclass cell_simple[id_t, real_t]:
        id_t m_son[1]
        real_t m_xg[3]
        id_t m_cpu

        char m_pos
        cell_simple()
        bint is_refined(int ison)

    # AMR level implementation

    # This class definition is out of date.  I have unrolled the template
    # below.
    cdef cppclass level[Cell_]:
        unsigned m_ilevel
        vector[Cell_] m_level_cells
        double m_xc[8]
        double m_yc[8]
        double m_zc[8]

        # I skipped the typedefs here

        double m_dx
        unsigned m_nx
        level (unsigned ilevel)

        void register_cell( Cell_ cell )
        vector[Cell_].iterator begin()
        vector[Cell_].iterator end()

        Cell_& operator[]( unsigned i)
        unsigned size()

    cdef cppclass RAMSES_level:
        unsigned m_ilevel
        vector[RAMSES_cell] m_level_cells
        double m_xc[8]
        double m_yc[8]
        double m_zc[8]

        # I skipped the typedefs here

        double m_dx
        unsigned m_nx
        RAMSES_level (unsigned ilevel)

        void register_cell( RAMSES_cell cell )
        vector[RAMSES_cell].iterator begin()
        vector[RAMSES_cell].iterator end()

        RAMSES_cell& operator[]( unsigned i)
        unsigned size()

    # This class definition is out of date.  I have unrolled the template
    # below.
    cdef cppclass tree[Cell_, Level_]:
        cppclass header:
            vector[int] nx
            vector[int] nout
            vector[int] nsteps
            int ncpu
            int ndim
            int nlevelmax
            int ngridmax
            int nboundary
            int ngrid_current
            double boxlen
            vector[double] tout
            vector[double] aout
            vector[double] dtold
            vector[double] dtnew
            vector[double] cosm
            vector[double] timing
            double t
            double mass_sph

        vector[Level_] m_AMR_levels
        vector[unsigned] mheadl, m_numbl, m_taill

        int m_cpu
        int m_minlevel
        int m_maxlevel
        string m_fname
        unsigned m_ncoarse
        header m_header

        # This is from later on in the .hh file ... I don't think we need them
        # typedef proto_iterator<const tree*> const_iterator;
        # typedef proto_iterator<tree *> iterator;

        tree (snapshot &snap, int cpu, int maxlevel, int minlevel = 1)

    cppclass tree_iterator "RAMSES::AMR::RAMSES_tree::iterator":
        tree_iterator operator*()
        RAMSES_cell operator*()
        tree_iterator begin()
        tree_iterator end()
        tree_iterator to_parent()
        tree_iterator get_parent()
        void next()
        bint operator!=(tree_iterator other)
        unsigned get_cell_father()
        bint is_finest(int ison)
        int get_absolute_position()
        int get_domain()

    cdef cppclass RAMSES_tree:
        # This is, strictly speaking, not a header.  But, I believe it is
        # going to work alright.
        cppclass header:
            vector[int] nx
            vector[int] nout
            vector[int] nsteps
            int ncpu
            int ndim
            int nlevelmax
            int ngridmax
            int nboundary
            int ngrid_current
            double boxlen
            vector[double] tout
            vector[double] aout
            vector[double] dtold
            vector[double] dtnew
            vector[double] cosm
            vector[double] timing
            double t
            double mass_sph

        vector[RAMSES_level] m_AMR_levels
        vector[unsigned] mheadl, m_numbl, m_taill

        int m_cpu
        int m_minlevel
        int m_maxlevel
        string m_fname
        unsigned m_ncoarse
        header m_header

        unsigned size()

        # This is from later on in the .hh file ... I don't think we need them
        # typedef proto_iterator<const tree*> const_iterator;
        # typedef proto_iterator<tree *> iterator;

        RAMSES_tree(snapshot &snap, int cpu, int maxlevel, int minlevel)
        void read()

        tree_iterator begin(int ilevel)
        tree_iterator end(int ilevel)

        tree_iterator begin()
        tree_iterator end()

        vec[double] cell_pos_double "cell_pos<double>" (tree_iterator it, unsigned ind) 
        vec[double] grid_pos_double "grid_pos<double>" (tree_iterator it)
        vec[float] cell_pos_float "cell_pos<float>" (tree_iterator it, unsigned ind) 
        vec[float] grid_pos_float "grid_pos<float>" (tree_iterator it)

cdef extern from "RAMSES_amr_data.hh" namespace "RAMSES::HYDRO":
    enum hydro_var:
        density
        velocity_x
        velocity_y
        velocity_z
        pressure
        metallicit

    char ramses_hydro_variables[][64]

    # There are a number of classes here that we could wrap and utilize.
    # However, I will only wrap the methods I know we need.

    # I have no idea if this will work.
    cdef cppclass TreeTypeIterator[TreeType_]:
        pass

    # This class definition is out of date.  I have unrolled the template
    # below.
    cdef cppclass data[TreeType_, Real_]:
        cppclass header:
            unsigned ncpu
            unsigned nvar
            unsigned ndim
            unsigned nlevelmax
            unsigned nboundary
            double gamma
        string m_fname
        header m_header

        # I don't want to implement proto_data, so we put this here
        Real_& cell_value(TreeTypeIterator[TreeType_] &it, int ind)

        unsigned m_nvars
        vector[string] m_varnames
        map[string, unsigned] m_var_name_map

        data(TreeType_ &AMRtree)
        #_OutputIterator get_var_names[_OutputIterator](_OutputIterator names)
        void read(string varname)

    cdef cppclass RAMSES_hydro_data:
        cppclass header:
            unsigned ncpu
            unsigned nvar
            unsigned ndim
            unsigned nlevelmax
            unsigned nboundary
            double gamma
        string m_fname
        header m_header

        # I don't want to implement proto_data, so we put this here
        double cell_value (tree_iterator &it, int ind)

        unsigned m_nvars
        vector[string] m_varnames
        map[string, unsigned] m_var_name_map

        RAMSES_hydro_data(RAMSES_tree &AMRtree)
        #_OutputIterator get_var_names[_OutputIterator](_OutputIterator names)
        void read(string varname)

        vector[vector[double]] m_var_array

cdef class RAMSES_tree_proxy:
    cdef string *snapshot_name
    cdef snapshot *rsnap

    cdef RAMSES_tree** trees
    cdef RAMSES_hydro_data*** hydro_datas

    cdef int **loaded

    cdef public object field_ind
    cdef public object field_names

    # We will store this here so that we have a record, independent of the
    # header, of how many things we have allocated
    cdef int ndomains, nfields
    
    def __cinit__(self, char *fn):
        cdef int idomain, ifield, ii
        cdef RAMSES_tree *local_tree
        cdef RAMSES_hydro_data *local_hydro_data
        self.snapshot_name = new string(fn)
        self.rsnap = new snapshot(deref(self.snapshot_name), version3)
        # We now have to get our field names to fill our array
        self.trees = <RAMSES_tree**>\
            malloc(sizeof(RAMSES_tree*) * self.rsnap.m_header.ncpu)
        for ii in range(self.ndomains): self.trees[ii] = NULL
        self.hydro_datas = <RAMSES_hydro_data ***>\
                       malloc(sizeof(RAMSES_hydro_data**) * self.rsnap.m_header.ncpu)
        self.ndomains = self.rsnap.m_header.ncpu
        
        # Note we don't do ncpu + 1
        for idomain in range(self.rsnap.m_header.ncpu):
            # we don't delete local_tree
            local_tree = new RAMSES_tree(deref(self.rsnap), idomain + 1,
                                         self.rsnap.m_header.levelmax, 0)
            local_tree.read()
            local_hydro_data = new RAMSES_hydro_data(deref(local_tree))
            self.hydro_datas[idomain] = <RAMSES_hydro_data **>\
                malloc(sizeof(RAMSES_hydro_data*) * local_hydro_data.m_nvars)
            for ii in range(local_hydro_data.m_nvars):
                self.hydro_datas[idomain][ii] = \
                    new RAMSES_hydro_data(deref(local_tree))
            self.trees[idomain] = local_tree
            # We do not delete the final snapshot, which we'll use later
            #if idomain + 1 < self.rsnap.m_header.ncpu:
            #    del local_hydro_data
        # Only once, we read all the field names
        self.nfields = local_hydro_data.m_nvars
        cdef string *field_name
        self.field_names = []
        self.field_ind = {}
        self.loaded = <int **> malloc(sizeof(int*) * self.ndomains)
        for idomain in range(self.ndomains):
            self.loaded[idomain] = <int *> malloc(
                sizeof(int) * local_hydro_data.m_nvars)
            for ifield in range(local_hydro_data.m_nvars):
                self.loaded[idomain][ifield] = 0
        for ifield in range(local_hydro_data.m_nvars):
            field_name = &(local_hydro_data.m_varnames[ifield])
            # Does this leak?
            self.field_names.append(field_name.c_str())
            self.field_ind[self.field_names[-1]] = ifield
        # This all needs to be cleaned up in the deallocator

    def get_domain_boundaries(self):
        bounds = []
        for i in range(self.rsnap.m_header.ncpu):
            bounds.append((self.rsnap.ind_min[i],
                           self.rsnap.ind_max[i]))
        return bounds

    def __dealloc__(self):
        cdef int idomain, ifield
        # To ensure that 'delete' is used, not 'free',
        # we allocate temporary variables.
        cdef RAMSES_tree *temp_tree
        cdef RAMSES_hydro_data *temp_hdata
        for idomain in range(self.ndomains):
            for ifield in range(self.nfields):
                temp_hdata = self.hydro_datas[idomain][ifield]
                del temp_hdata
            temp_tree = self.trees[idomain]
            del temp_tree
            free(self.loaded[idomain])
            free(self.hydro_datas[idomain])
        free(self.trees)
        free(self.hydro_datas)
        free(self.loaded)
        if self.snapshot_name != NULL: del self.snapshot_name
        if self.rsnap != NULL: del self.rsnap
        
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def count_zones(self):
        # We need to do simulation domains here

        cdef unsigned idomain, ilevel
        cdef RAMSES_tree *local_tree
        cdef RAMSES_hydro_data *local_hydro_data
        cdef RAMSES_level *local_level
        cdef tree_iterator grid_it, grid_end

        # All the loop-local pointers must be declared up here

        cdef np.ndarray[np.int64_t, ndim=1] cell_count
        cell_count = np.zeros(self.rsnap.m_header.levelmax + 1, 'int64')
        cdef int local_count = 0
        cdef int tree_count
        for idomain in range(1, self.rsnap.m_header.ncpu + 1):
            local_tree = self.trees[idomain - 1]
            for ilevel in range(local_tree.m_maxlevel + 1):
                local_count = 0
                tree_count = 0
                local_level = &local_tree.m_AMR_levels[ilevel]
                grid_it = local_tree.begin(ilevel)
                grid_end = local_tree.end(ilevel)
                while grid_it != grid_end:
                    local_count += (grid_it.get_domain() == idomain)
                    tree_count += 1
                    grid_it.next()
                cell_count[ilevel] += local_count

        return cell_count

    cdef ensure_loaded(self, char *varname, int domain_index, int varindex = -1):
        # this domain_index must be zero-indexed
        if varindex == -1: varindex = self.field_ind[varname]
        if self.loaded[domain_index][varindex] == 1:
            return
        cdef string *field_name = new string(varname)
        print "READING FROM DISK", varname, domain_index, varindex
        self.hydro_datas[domain_index][varindex].read(deref(field_name))
        self.loaded[domain_index][varindex] = 1
        del field_name

    def clear_tree(self, char *varname, int domain_index):
        # this domain_index must be zero-indexed
        # We delete and re-create
        cdef int varindex = self.field_ind[varname]
        cdef string *field_name = new string(varname)
        cdef RAMSES_hydro_data *temp_hdata
        if self.loaded[domain_index][varindex] == 1:
            temp_hdata = self.hydro_datas[domain_index][varindex]
            del temp_hdata
            self.hydro_datas[domain_index][varindex] = \
                new RAMSES_hydro_data(deref(self.trees[domain_index]))
            self.loaded[domain_index][varindex] = 0
        del field_name

    def get_file_info(self):
        header_info = {}
        header_info["ncpu"] = self.rsnap.m_header.ncpu
        header_info["ndim"] = self.rsnap.m_header.ndim
        header_info["levelmin"] = self.rsnap.m_header.levelmin
        header_info["levelmax"] = self.rsnap.m_header.levelmax
        header_info["ngridmax"] = self.rsnap.m_header.ngridmax
        header_info["nstep_coarse"] = self.rsnap.m_header.nstep_coarse
        header_info["boxlen"] = self.rsnap.m_header.boxlen
        header_info["time"] = self.rsnap.m_header.time
        header_info["aexp"] = self.rsnap.m_header.aexp
        header_info["H0"] = self.rsnap.m_header.H0
        header_info["omega_m"] = self.rsnap.m_header.omega_m
        header_info["omega_l"] = self.rsnap.m_header.omega_l
        header_info["omega_k"] = self.rsnap.m_header.omega_k
        header_info["omega_b"] = self.rsnap.m_header.omega_b
        header_info["unit_l"] = self.rsnap.m_header.unit_l
        header_info["unit_d"] = self.rsnap.m_header.unit_d
        header_info["unit_t"] = self.rsnap.m_header.unit_t

        # Now we grab some from the trees
        cdef np.ndarray[np.int32_t, ndim=1] top_grid_dims = np.zeros(3, "int32")
        cdef int i
        for i in range(3):
            # Note that nx is really boundary conditions.  We always have 2.
            top_grid_dims[i] = self.trees[0].m_header.nx[i]
            top_grid_dims[i] = 2
        header_info["nx"] = top_grid_dims

        return header_info

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def fill_hierarchy_arrays(self, 
                              np.ndarray[np.int32_t, ndim=1] top_grid_dims,
                              np.ndarray[np.float64_t, ndim=2] left_edges,
                              np.ndarray[np.float64_t, ndim=2] right_edges,
                              np.ndarray[np.int32_t, ndim=2] grid_levels,
                              np.ndarray[np.int64_t, ndim=2] grid_file_locations,
                              np.ndarray[np.uint64_t, ndim=1] grid_hilbert_indices,
                              np.ndarray[np.int32_t, ndim=2] child_mask,
                              np.ndarray[np.float64_t, ndim=1] domain_left,
                              np.ndarray[np.float64_t, ndim=1] domain_right):
        # We need to do simulation domains here

        cdef unsigned idomain, ilevel

        # All the loop-local pointers must be declared up here
        cdef RAMSES_tree *local_tree
        cdef RAMSES_hydro_data *local_hydro_data
        cdef unsigned father

        cdef tree_iterator grid_it, grid_end, father_it
        cdef vec[double] gvec
        cdef int grid_ind = 0, grid_aind = 0
        cdef unsigned parent_ind
        cdef bint ci
        cdef double pos[3]
        cdef double grid_half_width 
        cdef unsigned long rv

        cdef np.int32_t rr
        cdef int i
        cdef np.ndarray[np.int64_t, ndim=1] level_cell_counts
        level_cell_counts = np.zeros(self.rsnap.m_header.levelmax + 1, 'int64')
        for idomain in range(1, self.rsnap.m_header.ncpu + 1):
            local_tree = self.trees[idomain - 1]
            for ilevel in range(local_tree.m_maxlevel + 1):
                # this gets overwritten for every domain, which is okay
                level_cell_counts[ilevel] = grid_ind 
                #grid_half_width = self.rsnap.m_header.boxlen / \
                grid_half_width = 1.0 / \
                    (2**(ilevel) * top_grid_dims[0])
                grid_it = local_tree.begin(ilevel)
                grid_end = local_tree.end(ilevel)
                while grid_it != grid_end:
                    if grid_it.get_domain() != idomain:
                        grid_ind += 1
                        grid_it.next()
                        continue
                    gvec = local_tree.grid_pos_double(grid_it)
                    left_edges[grid_aind, 0] = pos[0] = gvec.x - grid_half_width
                    left_edges[grid_aind, 1] = pos[1] = gvec.y - grid_half_width
                    left_edges[grid_aind, 2] = pos[2] = gvec.z - grid_half_width
                    for i in range(3):
                        pos[i] = (pos[i] - domain_left[i]) / (domain_right[i] - domain_left[i])
                    right_edges[grid_aind, 0] = gvec.x + grid_half_width
                    right_edges[grid_aind, 1] = gvec.y + grid_half_width
                    right_edges[grid_aind, 2] = gvec.z + grid_half_width
                    grid_levels[grid_aind, 0] = ilevel
                    # Now the harder part
                    father_it = grid_it.get_parent()
                    grid_file_locations[grid_aind, 0] = <np.int64_t> idomain
                    grid_file_locations[grid_aind, 1] = grid_ind - level_cell_counts[ilevel]
                    parent_ind = father_it.get_absolute_position()
                    if ilevel > 0:
                        # We calculate the REAL parent index
                        grid_file_locations[grid_aind, 2] = \
                            level_cell_counts[ilevel - 1] + parent_ind
                    else:
                        grid_file_locations[grid_aind, 2] = -1
                    for ci in range(8):
                        rr = <np.int32_t> grid_it.is_finest(ci)
                        child_mask[grid_aind, ci] = rr
                    grid_ind += 1
                    grid_aind += 1
                    grid_it.next()

    def read_oct_grid(self, char *field, int level, int domain, int grid_id):

        self.ensure_loaded(field, domain - 1)
        cdef int varindex = self.field_ind[field]
        cdef int i

        cdef np.ndarray[np.float64_t, ndim=3] tr = np.empty((2,2,2), dtype='float64',
                                                   order='F')
        cdef tree_iterator grid_it, grid_end
        cdef double* data = <double*> tr.data

        cdef RAMSES_tree *local_tree = self.trees[domain - 1]
        cdef RAMSES_hydro_data *local_hydro_data = self.hydro_datas[domain - 1][varindex]

        #inline ValueType_& cell_value( const typename TreeType_::iterator& it,
        #                               int ind )
        #{
        #   unsigned ipos   = it.get_absolute_position();
        #   unsigned ilevel = it.get_level();//-m_minlevel;
        #   return (m_var_array[ilevel])[m_twotondim*ipos+ind];
        #}
        
        for i in range(8): 
            data[i] = local_hydro_data.m_var_array[level][8*grid_id+i]
        return tr

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def read_grid(self, int varindex, char *field,
                  np.ndarray[np.int64_t, ndim=1] start_index,
                  np.ndarray[np.int32_t, ndim=1] grid_dims,
                  np.ndarray[np.float64_t, ndim=3] data,
                  np.ndarray[np.int32_t, ndim=3] filled,
                  int level, int ref_factor,
                  np.ndarray[np.int64_t, ndim=2] component_grid_info):
        cdef RAMSES_tree *local_tree = NULL
        cdef RAMSES_hydro_data *local_hydro_data = NULL

        cdef int gi, i, j, k, domain, offset
        cdef int ir, jr, kr
        cdef int n
        cdef int offi, offj, offk, odind
        cdef np.int64_t di, dj, dk
        cdef np.int32_t og_start_index[3]
        cdef np.float64_t temp_data
        cdef np.int64_t end_index[3]
        cdef int to_fill = 0
        # Note that indexing into a cell is:
        #   (k*2 + j)*2 + i
        for i in range(3):
            end_index[i] = start_index[i] + grid_dims[i]
        for gi in range(component_grid_info.shape[0]):
            domain = component_grid_info[gi,0]
            if domain == 0: continue
            self.ensure_loaded(field, domain - 1, varindex)
            local_tree = self.trees[domain - 1]
            local_hydro_data = self.hydro_datas[domain - 1][varindex]
            offset = component_grid_info[gi,1]
            for n in range(3): og_start_index[n] = component_grid_info[gi,3+n]
            for i in range(2*ref_factor):
                di = i + og_start_index[0] * ref_factor
                if di < start_index[0] or di >= end_index[0]: continue
                ir = <int> (i / ref_factor)
                for j in range(2 * ref_factor):
                    dj = j + og_start_index[1] * ref_factor
                    if dj < start_index[1] or dj >= end_index[1]: continue
                    jr = <int> (j / ref_factor)
                    for k in range(2 * ref_factor):
                        dk = k + og_start_index[2] * ref_factor
                        if dk < start_index[2] or dk >= end_index[2]: continue
                        kr = <int> (k / ref_factor)
                        offi = di - start_index[0]
                        offj = dj - start_index[1]
                        offk = dk - start_index[2]
                        #print offi, filled.shape[0],
                        #print offj, filled.shape[1],
                        #print offk, filled.shape[2]
                        if filled[offi, offj, offk] == 1: continue

                        odind = (kr*2 + jr)*2 + ir
                        temp_data = local_hydro_data.m_var_array[
                                level][8*offset + odind]
                        data[offi, offj, offk] = temp_data
                        filled[offi, offj, offk] = 1
                        to_fill += 1
        return to_fill

cdef class ProtoSubgrid:
    cdef np.int64_t *signature[3]
    cdef np.int64_t left_edge[3]
    cdef np.int64_t right_edge[3]
    cdef np.int64_t dimensions[3]
    cdef public np.float64_t efficiency
    cdef np.int64_t *sigs[3]
    cdef public object grid_file_locations
    cdef public object dd
        
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __cinit__(self,
                   np.ndarray[np.int64_t, ndim=1] left_index,
                   np.ndarray[np.int64_t, ndim=1] dimensions, 
                   np.ndarray[np.int64_t, ndim=2] left_edges,
                   np.ndarray[np.int64_t, ndim=2] grid_file_locations):
        # This also includes the shrinking step.
        cdef int i, ci, ng = left_edges.shape[0]
        cdef np.ndarray temp_arr
        cdef int l0, r0, l1, r1, l2, r2, i0, i1, i2
        cdef np.int64_t temp_l[3], temp_r[3], ncells
        cdef np.float64_t efficiency
        for i in range(3):
            temp_l[i] = left_index[i] + dimensions[i]
            temp_r[i] = left_index[i]
            self.signature[i] = NULL
        for gi in range(ng):
            if left_edges[gi,0] > left_index[0]+dimensions[0] or \
               left_edges[gi,0] + 2 < left_index[0] or \
               left_edges[gi,1] > left_index[1]+dimensions[1] or \
               left_edges[gi,1] + 2 < left_index[1] or \
               left_edges[gi,2] > left_index[2]+dimensions[2] or \
               left_edges[gi,2] + 2 < left_index[2]:
               #print "Skipping grid", gi, "which lies outside out box"
               continue
            for i in range(3):
                temp_l[i] = i64min(left_edges[gi,i], temp_l[i])
                temp_r[i] = i64max(left_edges[gi,i] + 2, temp_r[i])
        for i in range(3):
            self.left_edge[i] = i64max(temp_l[i], left_index[i])
            self.right_edge[i] = i64min(temp_r[i], left_index[i] + dimensions[i])
            self.dimensions[i] = self.right_edge[i] - self.left_edge[i]
            if self.dimensions[i] <= 0:
                self.efficiency = -1.0
                return
            self.sigs[i] = <np.int64_t *> malloc(
                                sizeof(np.int64_t) * self.dimensions[i])
            for gi in range(self.dimensions[i]): self.sigs[i][gi] = 0
        
        # My guess is that this whole loop could be done more efficiently.
        # However, this is clear and straightforward, so it is a good first
        # pass.
        cdef np.int64_t *sig0, *sig1, *sig2
        sig0 = self.sigs[0]
        sig1 = self.sigs[1]
        sig2 = self.sigs[2]
        efficiency = 0.0
        cdef int used
        cdef np.ndarray[np.int32_t, ndim=1] mask
        mask = np.zeros(ng, 'int32')
        used = 0
        for gi in range(ng):
            for l0 in range(2):
                i0 = left_edges[gi, 0] + l0
                if i0 < self.left_edge[0]: continue
                if i0 >= self.right_edge[0]: break
                for l1 in range(2):
                    i1 = left_edges[gi, 1] + l1
                    if i1 < self.left_edge[1]: continue
                    if i1 >= self.right_edge[1]: break
                    for l2 in range(2):
                        i2 = left_edges[gi, 2] + l2
                        if i2 < self.left_edge[2]: continue
                        if i2 >= self.right_edge[2]: break
                        i = i0 - self.left_edge[0]
                        sig0[i] += 1
                        i = i1 - self.left_edge[1]
                        sig1[i] += 1
                        i = i2 - self.left_edge[2]
                        sig2[i] += 1
                        efficiency += 1
                        mask[gi] = 1
            used += mask[gi]
        cdef np.ndarray[np.int64_t, ndim=2] gfl
        gfl = np.zeros((used, 6), 'int64')
        used = 0
        self.grid_file_locations = gfl
        for gi in range(ng):
            if mask[gi] == 1:
                grid_file_locations[gi,3] = left_edges[gi,0]
                grid_file_locations[gi,4] = left_edges[gi,1]
                grid_file_locations[gi,5] = left_edges[gi,2]
                for i in range(6):
                    gfl[used, i] = grid_file_locations[gi,i]
                used += 1
         
        self.dd = np.ones(3, dtype='int64')
        for i in range(3):
            efficiency /= self.dimensions[i]
            self.dd[i] = self.dimensions[i]
        #print "Efficiency is %0.3e" % (efficiency)
        self.efficiency = efficiency

    def __dealloc__(self):
        free(self.sigs[0])
        free(self.sigs[1])
        free(self.sigs[2])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void find_split(self, int *tr,):
        # First look for zeros
        cdef int i, center, ax
        cdef np.ndarray[ndim=1, dtype=np.int64_t] axes
        cdef np.int64_t strength, zcstrength, zcp
        axes = np.argsort(self.dd)[::-1]
        cdef np.int64_t *sig
        for axi in range(3):
            ax = axes[axi] #iterate over domain dimensions
            center = self.dimensions[ax] / 2 
            sig = self.sigs[ax] #an array for the dimension, number of cells along that dim
            for i in range(self.dimensions[ax]):
                if sig[i] == 0 and i > 0 and i < self.dimensions[ax] - 1:
                    #print "zero: %s (%s)" % (i, self.dimensions[ax])
                    tr[0] = 0; tr[1] = ax; tr[2] = i
                    return
        zcstrength = 0
        zcp = 0
        zca = -1
        cdef int temp
        cdef np.int64_t *sig2d
        for axi in range(3):
            ax = axes[axi]
            sig = self.sigs[ax]
            sig2d = <np.int64_t *> malloc(sizeof(np.int64_t) * self.dimensions[ax])
            sig2d[0] = sig2d[self.dimensions[ax]-1] = 0
            for i in range(1, self.dimensions[ax] - 1):
                sig2d[i] = sig[i-1] - 2*sig[i] + sig[i+1]
            for i in range(1, self.dimensions[ax] - 1):
                if sig2d[i] * sig2d[i+1] <= 0:
                    strength = labs(sig2d[i] - sig2d[i+1])
                    if (strength > zcstrength) or \
                       (strength == zcstrength and (abs(center - i) <
                                                    abs(center - zcp))):
                        zcstrength = strength
                        zcp = i
                        zca = ax
            free(sig2d)
        #print "zcp: %s (%s)" % (zcp, self.dimensions[ax])
        tr[0] = 1; tr[1] = ax; tr[2] = zcp
        return

    @cython.wraparound(False)
    cdef void find_split_center(self, int *tr,):
        # First look for zeros
        cdef int i, center, ax
        cdef int flip
        cdef np.ndarray[ndim=1, dtype=np.int64_t] axes
        cdef np.int64_t strength, zcstrength, zcp
        axes = np.argsort(self.dd)[::-1]
        cdef np.int64_t *sig
        for axi in range(3):
            ax = axes[axi] #iterate over domain dimensions
            center = self.dimensions[ax] / 2 
            sig = self.sigs[ax] #an array for the dimension, number of cells along that dim
            #frequently get stuck with many zeroes near the edge of the grid
            #let's start from the middle, working out
            for j in range(self.dimensions[ax]/2):
                flip = 1
                i = self.dimensions[ax]/2+j
                if sig[i] == 0 and i > 0 and i < self.dimensions[ax] - 1:
                    #print "zero: %s (%s)" % (i, self.dimensions[ax])
                    tr[0] = 0; tr[1] = ax; tr[2] = i
                    return
                i = self.dimensions[ax]/2-j
                if sig[i] == 0 and i > 0 and i < self.dimensions[ax] - 1:
                    #print "zero: %s (%s)" % (i, self.dimensions[ax])
                    tr[0] = 0; tr[1] = ax; tr[2] = i
                    return
                    
                
        zcstrength = 0
        zcp = 0
        zca = -1
        cdef int temp
        cdef np.int64_t *sig2d
        for axi in range(3):
            ax = axes[axi]
            sig = self.sigs[ax]
            sig2d = <np.int64_t *> malloc(sizeof(np.int64_t) * self.dimensions[ax])
            sig2d[0] = sig2d[self.dimensions[ax]-1] = 0
            for i in range(1, self.dimensions[ax] - 1):
                sig2d[i] = sig[i-1] - 2*sig[i] + sig[i+1]
            for i in range(1, self.dimensions[ax] - 1):
                if sig2d[i] * sig2d[i+1] <= 0:
                    strength = labs(sig2d[i] - sig2d[i+1])
                    if (strength > zcstrength) or \
                       (strength == zcstrength and (abs(center - i) <
                                                    abs(center - zcp))):
                        zcstrength = strength
                        zcp = i
                        zca = ax
            free(sig2d)
        #print "zcp: %s (%s)" % (zcp, self.dimensions[ax])
        tr[0] = 1; tr[1] = ax; tr[2] = zcp
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_properties(self):
        cdef np.ndarray[np.int64_t, ndim=2] tr = np.empty((3,3), dtype='int64')
        cdef int i
        for i in range(3):
            tr[0,i] = self.left_edge[i]
            tr[1,i] = self.right_edge[i]
            tr[2,i] = self.dimensions[i]
        return tr

@cython.cdivision
cdef np.int64_t graycode(np.int64_t x):
    return x^(x>>1)

@cython.cdivision
cdef np.int64_t igraycode(np.int64_t x):
    cdef np.int64_t i, j
    if x == 0:
        return x
    m = <np.int64_t> ceil(log2(x)) + 1
    i, j = x, 1
    while j < m:
        i = i ^ (x>>j)
        j += 1
    return i

@cython.cdivision
cdef np.int64_t direction(np.int64_t x, np.int64_t n):
    #assert x < 2**n
    if x == 0:
        return 0
    elif x%2 == 0:
        return tsb(x-1, n)%n
    else:
        return tsb(x, n)%n

@cython.cdivision
cdef np.int64_t tsb(np.int64_t x, np.int64_t width):
    #assert x < 2**width
    cdef np.int64_t i = 0
    while x&1 and i <= width:
        x = x >> 1
        i += 1
    return i

@cython.cdivision
cdef np.int64_t bitrange(np.int64_t x, np.int64_t width,
                         np.int64_t start, np.int64_t end):
    return x >> (width-end) & ((2**(end-start))-1)

@cython.cdivision
cdef np.int64_t rrot(np.int64_t x, np.int64_t i, np.int64_t width):
    i = i%width
    x = (x>>i) | (x<<width-i)
    return x&(2**width-1)

@cython.cdivision
cdef np.int64_t lrot(np.int64_t x, np.int64_t i, np.int64_t width):
    i = i%width
    x = (x<<i) | (x>>width-i)
    return x&(2**width-1)

@cython.cdivision
cdef np.int64_t transform(np.int64_t entry, np.int64_t direction,
                          np.int64_t width, np.int64_t x):
    return rrot((x^entry), direction + 1, width)

@cython.cdivision
cdef np.int64_t entry(np.int64_t x):
    if x == 0: return 0
    return graycode(2*((x-1)/2))

@cython.cdivision
def get_hilbert_indices(int order, np.ndarray[np.int64_t, ndim=2] left_index):
    # This is inspired by the scurve package by user cortesi on GH.
    cdef int o, i
    cdef np.int64_t x, w, h, e, d, l, b
    cdef np.int64_t p[3]
    cdef np.ndarray[np.int64_t, ndim=1] hilbert_indices
    hilbert_indices = np.zeros(left_index.shape[0], 'int64')
    for o in range(left_index.shape[0]):
        p[0] = left_index[o, 0]
        p[1] = left_index[o, 1]
        p[2] = left_index[o, 2]
        h = e = d = 0
        for i in range(order):
            l = 0
            for x in range(3):
                b = bitrange(p[3-x-1], order, i, i+1)
                l |= (b<<x)
            l = transform(e, d, 3, l)
            w = igraycode(l)
            e = e ^ lrot(entry(w), d+1, 3)
            d = (d + direction(w, 3) + 1)%3
            h = (h<<3)|w
        hilbert_indices[o] = h
    return hilbert_indices


@cython.boundscheck(False)
@cython.wraparound(False)
def get_array_indices_lists(np.ndarray[np.int64_t, ndim=1] ind, 
                            np.ndarray[np.int64_t, ndim=1] uind,
                            np.ndarray[np.int64_t, ndim=2] lefts,
                            np.ndarray[np.int64_t, ndim=2] files):
    #ind are the hilbert indices 
    #uind are the unique hilbert indices                        
    #count[n] track of how many times the nth index of uind occurs in ind
    
    cdef np.ndarray[np.int64_t, ndim=1] count = np.zeros(uind.shape[0], 'int64')
    cdef int n, i
    cdef np.int64_t mi, mui
    
    #fill in the count array
    for i in range(ind.shape[0]):
        mi = ind[i]
        for n in range(uind.shape[0]):
            if uind[n] == mi:
                count[n] += 1
                break
    
    cdef np.int64_t **alefts
    cdef np.int64_t **afiles
    afiles = <np.int64_t **> malloc(sizeof(np.int64_t *) * uind.shape[0])
    alefts = <np.int64_t **> malloc(sizeof(np.int64_t *) * uind.shape[0])
    cdef int *li = <int *> malloc(sizeof(int) * uind.shape[0])
    cdef np.ndarray[np.int64_t, ndim=2] locations
    cdef np.ndarray[np.int64_t, ndim=2] left
    all_locations = []
    all_lefts = []
    
    #having measure the repetition of each hilbert index,
    #we can know declare how much memory we will use
    for n in range(uind.shape[0]):
        locations = np.zeros((count[n], 6), 'int64')
        left = np.zeros((count[n], 3), 'int64')
        all_locations.append(locations)
        all_lefts.append(left)
        afiles[n] = <np.int64_t *> locations.data
        alefts[n] = <np.int64_t *> left.data
        li[n] = 0
    
    cdef int fi
    #now arrange all_locations and all_lefts sequentially
    #such that when they return to python
    #the 1d array mutates into a list of lists?
    for i in range(ind.shape[0]):
        mi = ind[i]
        for n in range(uind.shape[0]):
            if uind[n] == mi:
                for fi in range(3):
                    alefts[n][li[n] * 3 + fi] = lefts[i, fi]
                for fi in range(6):
                    afiles[n][li[n] * 6 + fi] = files[i, fi]
                li[n] += 1
                break
    free(afiles)
    free(alefts)
    return all_locations, all_lefts

def recursive_patch_splitting(ProtoSubgrid psg,
        np.ndarray[np.int64_t, ndim=1] dims,
        np.ndarray[np.int64_t, ndim=1] ind,
        np.ndarray[np.int64_t, ndim=2] left_index,
        np.ndarray[np.int64_t, ndim=2] fl,
        int num_deep = 0,
        float min_eff = 0.1,
        int use_center=0,
        long split_on_vol = 0):
    cdef ProtoSubgrid L, R
    cdef np.ndarray[np.int64_t, ndim=1] dims_l, li_l
    cdef np.ndarray[np.int64_t, ndim=1] dims_r, li_r
    cdef int tt, ax, fp, i, j, k, gi
    cdef int tr[3]
    cdef long volume  =0
    cdef int max_depth = 40
    volume = dims[0]*dims[1]*dims[2]
    if split_on_vol>0:
        if volume < split_on_vol:
            return [psg]
    if num_deep > max_depth:
        psg.efficiency = min_eff
        return [psg]
    if (psg.efficiency > min_eff or psg.efficiency < 0.0):
        return [psg]
    if not use_center:    
        psg.find_split(tr) #default
    else:
        psg.find_split_center(tr)    
        
    tt = tr[0]
    ax = tr[1]
    fp = tr[2]
    if (fp % 2) != 0:
        if dims[ax] != fp + 1:
            fp += 1
        else:
            fp -= 1
    dims_l = dims.copy()
    dims_l[ax] = fp
    li_l = ind.copy()
    for i in range(3):
        if dims_l[i] <= 0: return [psg]
    dims_r = dims.copy()
    dims_r[ax] -= fp
    li_r = ind.copy()
    li_r[ax] += fp
    for i in range(3):
        if dims_r[i] <= 0: return [psg]
    L = ProtoSubgrid(li_l, dims_l, left_index, fl)
    if L.efficiency > 1.0: raise RuntimeError
    if L.efficiency <= 0.0: rv_l = []
    elif L.efficiency < min_eff:
        rv_l = recursive_patch_splitting(L, dims_l, li_l,
                left_index, fl, num_deep + 1, min_eff,use_center,split_on_vol)
    else:
        rv_l = [L]
    R = ProtoSubgrid(li_r, dims_r, left_index, fl)
    if R.efficiency > 1.0: raise RuntimeError
    if R.efficiency <= 0.0: rv_r = []
    elif R.efficiency < min_eff:
        rv_r = recursive_patch_splitting(R, dims_r, li_r,
                left_index, fl, num_deep + 1, min_eff,use_center,split_on_vol)
    else:
        rv_r = [R]
    return rv_r + rv_l



"""
Author: Britton Smith <brittonsmith@gmail.com>
Affiliation:  UC Boulder, KIPAC
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
License:
  Copyright (C) 2008-2011 Britton Smith, Matthew Turk  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np

import matplotlib
import matplotlib.colors as cc
import matplotlib.cm as mcm
import _colormap_data as _cm

def is_colormap(cmap):
    return isinstance(cmap,cc.Colormap)

def check_color(name):
    try:
        ss = cc.colorConverter.to_rgb(name)
        return True
    except ValueError:
        return False

yt_colormaps = {}

def add_cmap(name, cdict):
    yt_colormaps[name] = \
        cc.LinearSegmentedColormap(name,cdict,256)
    mcm.datad[name] = cdict
    mcm.__dict__[name] = cdict
    try: # API compatibility
        mcm.register_cmap(name, yt_colormaps[name])
    except AttributeError:
        pass
    

# The format is as follows:
#   First number is the number at which we are defining a color breakpoint
#   Second number is the (0..1) number to interpolate to when coming *from below*
#   Third number is the (0..1) number to interpolate to when coming *from above*

# Next up is boilerplate -- the name, the colormap dict we just made, and the
# number of segments we want.  This is probably fine as is.

cdict = {'red':   ((0.0, 80/256., 80/256.),
                   (0.2, 0.0, 0.0),
                   (0.4, 0.0, 0.0),
                   (0.6, 256/256., 256/256.),
                   (0.95, 256/256., 256/256.),
                   (1.0, 150/256., 150/256.)),
         'green': ((0.0, 0/256., 0/256.),
                   (0.2, 0/256., 0/256.),
                   (0.4, 130/256., 130/256.),
                   (0.6, 256/256., 256/256.),
                   (1.0, 0.0, 0.0)),
         'blue':  ((0.0, 80/256., 80/256.),
                   (0.2, 220/256., 220/256.),
                   (0.4, 0.0, 0.0),
                   (0.6, 20/256., 20/256.),
                   (1.0, 0.0, 0.0))}

add_cmap('bds_highcontrast', cdict)
add_cmap('algae', cdict)

# Set the default colormap to be algae.
matplotlib.rc('image', cmap="algae")

# This next colormap was designed by Tune Kamae and converted here by Matt
_vs = np.linspace(0,1,255)
_kamae_red = np.minimum(255,
                113.9*np.sin(7.64*(_vs**1.705)+0.701)-916.1*(_vs+1.755)**1.862 \
              + 3587.9*_vs+2563.4)/255.0
_kamae_grn = np.minimum(255,
                70.0*np.sin(8.7*(_vs**1.26)-2.418)+151.7*_vs**0.5+70.0)/255.0
_kamae_blu = np.minimum(255,
                194.5*_vs**2.88+99.72*np.exp(-77.24*(_vs-0.742)**2.0)
              + 45.40*_vs**0.089+10.0)/255.0

cdict = {'red':zip(_vs,_kamae_red,_kamae_red),
         'green':zip(_vs,_kamae_grn,_kamae_grn),
         'blue':zip(_vs,_kamae_blu,_kamae_blu)}
add_cmap('kamae', cdict)

# This one is a simple black & green map

cdict = {'red':   ((0.0, 0.0, 0.0),
                   (1.0, 0.0, 0.0)),
         'green': ((0.0, 0.0, 0.0),
                   (1.0, 1.0, 1.0)),
         'blue':  ((0.0, 0.0, 0.0),
                   (1.0, 0.0, 0.0))}

add_cmap('black_green', cdict)

# This one comes from
# http://permalink.gmane.org/gmane.comp.python.matplotlib.devel/10518
# and is an implementation of http://arxiv.org/abs/1108.5083
#

# cubehelix parameters
_gamma_cubehelix = 1.0
_s_cubehelix = 0.5
_r_cubehelix = -1.5
_h_cubehelix = 1.0

_cubehelix_data = {
        'red': lambda x: x**_gamma_cubehelix + (_h_cubehelix * x**_gamma_cubehelix * (1 - x**_gamma_cubehelix) / 2) * (-0.14861 * np.cos(2 * np.pi * (_s_cubehelix / 3 + _r_cubehelix * x)) + 1.78277 * np.sin(2 * np.pi * (_s_cubehelix / 3 + _r_cubehelix * x))),
        'green': lambda x: x**_gamma_cubehelix + (_h_cubehelix * x**_gamma_cubehelix * (1 - x**_gamma_cubehelix) / 2) * (-0.29227 * np.cos(2 * np.pi * (_s_cubehelix / 3 + _r_cubehelix * x)) - 0.90649 * np.sin(2 * np.pi * (_s_cubehelix / 3 + _r_cubehelix * x))),
        'blue': lambda x: x**_gamma_cubehelix + (_h_cubehelix * x**_gamma_cubehelix * (1 - x**_gamma_cubehelix) / 2) * (1.97294 * np.cos(2 * np.pi * (_s_cubehelix / 3 + _r_cubehelix * x))),
}

add_cmap("cubehelix", _cubehelix_data)

# Add colormaps in _colormap_data.py that weren't defined here
_vs = np.linspace(0,1,255)
for k,v in _cm.color_map_luts.iteritems():
    if k not in yt_colormaps:
        cdict = { 'red': zip(_vs,v[0],v[0]),
                  'green': zip(_vs,v[1],v[1]),
                  'blue': zip(_vs,v[2],v[2]) }
        add_cmap(k, cdict)

def _extract_lookup_table(cmap_name):
    cmap = mcm.get_cmap(cmap_name)
    if not cmap._isinit: cmap._init()
    r = cmap._lut[:-3, 0]
    g = cmap._lut[:-3, 1]
    b = cmap._lut[:-3, 2]
    a = np.ones(b.shape)
    return [r, g, b, a]



#!/usr/bin/env python
import setuptools
import os, sys, os.path, glob, \
    tempfile, subprocess, shutil
from yt.utilities.setup import \
    check_for_dependencies


def check_for_png():
    return check_for_dependencies("PNG_DIR", "png.cfg", "png.h", "png")


def check_for_freetype():
    return check_for_dependencies(
        "FTYPE_DIR", "freetype.cfg", "ft2build.h", "freetype"
    )


def check_for_openmp():
    # Create a temporary directory
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    # Get compiler invocation
    compiler = os.getenv('CC', 'cc')

    # Attempt to compile a test script.
    # See http://openmp.org/wp/openmp-compilers/
    filename = r'test.c'
    file = open(filename,'w', 0)
    file.write(
        "#include <omp.h>\n"
        "#include <stdio.h>\n"
        "int main() {\n"
        "#pragma omp parallel\n"
        "printf(\"Hello from thread %d, nthreads %d\\n\", omp_get_thread_num(), omp_get_num_threads());\n"
        "}"
        )
    with open(os.devnull, 'w') as fnull:
        exit_code = subprocess.call([compiler, '-fopenmp', filename],
                                    stdout=fnull, stderr=fnull)

    # Clean up
    file.close()
    os.chdir(curdir)
    shutil.rmtree(tmpdir)

    if exit_code == 0:
        return True
    else:
        return False

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('lib',parent_package,top_path)
    png_inc, png_lib = check_for_png()
    freetype_inc, freetype_lib = check_for_freetype()
    if check_for_openmp() == True:
        omp_args = ['-fopenmp']
    else:
        omp_args = None
    # Because setjmp.h is included by lots of things, and because libpng hasn't
    # always properly checked its header files (see
    # https://bugzilla.redhat.com/show_bug.cgi?id=494579 ) we simply disable
    # support for setjmp.
    config.add_extension("CICDeposit", 
                ["yt/utilities/lib/CICDeposit.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("ContourFinding", 
                ["yt/utilities/lib/ContourFinding.pyx",
                 "yt/utilities/lib/union_find.c"],
                include_dirs=["yt/utilities/lib/"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("DepthFirstOctree", 
                ["yt/utilities/lib/DepthFirstOctree.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("fortran_reader", 
                ["yt/utilities/lib/fortran_reader.pyx"],
                include_dirs=["yt/utilities/lib/"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("freetype_writer", 
                ["yt/utilities/lib/freetype_writer.pyx"],
                include_dirs = [freetype_inc,os.path.join(freetype_inc, "freetype2"),
                                "yt/utilities/lib"],
                library_dirs = [freetype_lib], libraries=["freetype"],
                depends=["yt/utilities/lib/freetype_includes.h"])
    config.add_extension("geometry_utils", 
                ["yt/utilities/lib/geometry_utils.pyx"],
               extra_compile_args=omp_args,
               extra_link_args=omp_args,
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("Interpolators", 
                ["yt/utilities/lib/Interpolators.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("marching_cubes", 
                ["yt/utilities/lib/marching_cubes.pyx",
                 "yt/utilities/lib/FixedInterpolator.c"],
                include_dirs=["yt/utilities/lib/"],
                libraries=["m"],
                depends=["yt/utilities/lib/fp_utils.pxd",
                         "yt/utilities/lib/fixed_interpolator.pxd",
                         "yt/utilities/lib/FixedInterpolator.h",
                ])
    config.add_extension("misc_utilities", 
                ["yt/utilities/lib/misc_utilities.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("Octree", 
                ["yt/utilities/lib/Octree.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("png_writer", 
                ["yt/utilities/lib/png_writer.pyx"],
                define_macros=[("PNG_SETJMP_NOT_SUPPORTED", True)],
                include_dirs=[png_inc],
                library_dirs=[png_lib],
                libraries=["m", "png"],
                depends=["yt/utilities/lib/fp_utils.pxd"]),
    config.add_extension("PointsInVolume", 
                ["yt/utilities/lib/PointsInVolume.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("QuadTree", 
                ["yt/utilities/lib/QuadTree.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("RayIntegrators", 
                ["yt/utilities/lib/RayIntegrators.pyx"],
                libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("VolumeIntegrator", 
               ["yt/utilities/lib/VolumeIntegrator.pyx",
                "yt/utilities/lib/FixedInterpolator.c",
                "yt/utilities/lib/kdtree.c"] +
                 glob.glob("yt/utilities/lib/healpix_*.c"), 
               include_dirs=["yt/utilities/lib/"],
               libraries=["m"], 
               depends = ["yt/utilities/lib/VolumeIntegrator.pyx",
                          "yt/utilities/lib/fp_utils.pxd",
                          "yt/utilities/lib/healpix_interface.pxd",
                          "yt/utilities/lib/endian_swap.h",
                          "yt/utilities/lib/FixedInterpolator.h",
                          "yt/utilities/lib/healpix_vectors.h",
                          "yt/utilities/lib/kdtree.h",
                          "yt/utilities/lib/healpix_ang2pix_nest.c",
                          "yt/utilities/lib/healpix_mk_pix2xy.c",
                          "yt/utilities/lib/healpix_mk_xy2pix.c",
                          "yt/utilities/lib/healpix_pix2ang_nest.c",
                          "yt/utilities/lib/healpix_pix2vec_nest.c",
                          "yt/utilities/lib/healpix_vec2pix_nest.c"]
          )
    config.add_extension("grid_traversal", 
               ["yt/utilities/lib/grid_traversal.pyx",
                "yt/utilities/lib/FixedInterpolator.c",
                "yt/utilities/lib/kdtree.c"] +
                 glob.glob("yt/utilities/lib/healpix_*.c"), 
               include_dirs=["yt/utilities/lib/"],
               libraries=["m"], 
               extra_compile_args=omp_args,
               extra_link_args=omp_args,
               depends = ["yt/utilities/lib/VolumeIntegrator.pyx",
                          "yt/utilities/lib/fp_utils.pxd",
                          "yt/utilities/lib/kdtree.h",
                          "yt/utilities/lib/FixedInterpolator.h",
                          "yt/utilities/lib/fixed_interpolator.pxd",
                          "yt/utilities/lib/field_interpolation_tables.pxd",
                          ]
          )
    config.add_extension("write_array",
                         ["yt/utilities/lib/write_array.pyx"])
    config.add_extension("GridTree", 
    ["yt/utilities/lib/GridTree.pyx"],
        libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_extension("amr_kdtools", 
                         ["yt/utilities/lib/amr_kdtools.pyx"],
                         libraries=["m"], depends=["yt/utilities/lib/fp_utils.pxd"])
    config.add_subpackage("tests")

    if os.environ.get("GPERFTOOLS", "no").upper() != "NO":
        gpd = os.environ["GPERFTOOLS"]
        idir = os.path.join(gpd, "include")
        ldir = os.path.join(gpd, "lib")
        print "INCLUDE AND LIB DIRS", idir, ldir
        config.add_extension("perftools_wrap",
                ["yt/utilities/lib/perftools_wrap.pyx"],
                libraries=["profiler"],
                library_dirs = [ldir],
                include_dirs = [idir],
            )
    config.make_config_py() # installs __config__.py
    return config
